# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools

from oslo_log import log as logging


# inspired from neutron.log
def log(method, level=logging.DEBUG):
    """Decorator helping to log method calls."""

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        instance = args[0]
        data = {"class_name": instance.__class__.__name__,
                "method_name": method.__name__,
                "args": args[1:], "kwargs": kwargs}
        if hasattr(args[0], 'log'):
            logger = args[0].log
        else:
            logger = logging.getLogger(method.__module__)
        logger.log(level, 'method %(class_name)s.%(method_name)s'
                   ' called with %(args)s %(kwargs)s', data)
        return method(*args, **kwargs)
    return wrapper


def log_info(method):
    return log(method, logging.INFO)



# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import socket

import netaddr
from oslo_config import cfg
from oslo_serialization import jsonutils
import pyroute2
from pyroute2 import common as pr_common  # pylint: disable=no-name-in-module
from pyroute2 import netlink  # pylint: disable=no-name-in-module

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp import constants as consts
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.vpn import dataplane_drivers as dp_drivers
from networking_bagpipe.privileged import privileged_utils


ipr = pyroute2.IPRoute()  # pylint: disable=no-member

VRF_INTERFACE_PREFIX = "bvrf-"

RT_TABLE_BASE = 1000

RT_PROT_BAGPIPE = 19


# NOTE(tmorin): can this be removed ? (is jsonutils to_primitive() enough?)
def json_set_default(obj):
    if isinstance(obj, set):
        return list(obj)
    raise TypeError


def sysctl(sysctl_path, val):
    privileged_utils.sysctl(knob='.'.join(sysctl_path,), value=val)


def proxy_arp(ifname, enable):
    sysctl(['net', 'ipv4', 'conf', ifname, 'proxy_arp'], int(enable))
    sysctl(['net', 'ipv4', 'conf', ifname, 'proxy_arp_pvlan'], int(enable))


class MPLSLinuxVRFDataplane(dp_drivers.VPNInstanceDataplane):

    def __init__(self, *args, **kwargs):
        super(MPLSLinuxVRFDataplane, self).__init__(*args, **kwargs)

        # FIXME: maybe not thread safe ?
        self.ip = self.driver.ip

        self.vrf_if = ("%s%d" % (VRF_INTERFACE_PREFIX,
                                 self.instance_id))[:consts.LINUX_DEV_LEN]

        self.rt_table = RT_TABLE_BASE + self.instance_id

        self.log.info("Initializing VRF interface %s", self.vrf_if)

        self.flush_routes()

        self.log.debug("Creating VRF interface...")

        # Create VRF interface
        with self.ip.create(kind='vrf',
                            ifname=self.vrf_if,
                            vrf_table=self.rt_table,
                            reuse=True) as i:
            i.up()

        self.vrf_if_idx = i.index

        # Create ip rule for VRF route table
        # TODO(tmorin): do in IPDB, when possible (check: what if
        #               rule exist, but is not known by IPDB?)
        for family in socket.AF_INET, socket.AF_INET6:
            ipr.flush_rules(family=family,
                            iifname=self.vrf_if)
            ipr.flush_rules(family=family,
                            oifname=self.vrf_if)

            ipr.rule('add', family=family,
                     priority=100, iifname=self.vrf_if, table=self.rt_table,
                     action='FR_ACT_TO_TBL')
            ipr.rule('add', family=family,
                     priority=100, oifname=self.vrf_if, table=self.rt_table,

                     action='FR_ACT_TO_TBL')

            # if VRF traffic does not match any route,
            # lookups must *not* fallback to main/default
            # routing table
            ipr.rule('add', family=family,
                     priority=101, iifname=self.vrf_if,
                     action='FR_ACT_UNREACHABLE')
            ipr.rule('add', family=family,
                     priority=101, oifname=self.vrf_if,
                     action='FR_ACT_UNREACHABLE')

    def add_route(self, route):
        route.update({'proto': RT_PROT_BAGPIPE})
        self.ip.routes.add(route).commit()

    def flush_routes(self):
        ipr.flush_routes(table=self.rt_table)

    @log_decorator.log_info
    def cleanup(self):
        # bring down and disconnect all interfaces from vrf interface
        with self.ip.interfaces[self.vrf_if_idx] as vrf:
            for interface in vrf.ports:
                with self.ip.interfaces[interface] as i:
                    i.down()
                vrf.del_port(interface)
            vrf.remove()
        ipr.flush_rules(iifname=self.vrf_if)
        ipr.flush_rules(oifname=self.vrf_if)
        self.flush_routes()

    @log_decorator.log_info
    def vif_plugged(self, mac_address, ip_address, localport, label,
                    direction):

        interface = localport['linuxif']

        if interface not in self.ip.interfaces:
            self.log.warning("interface %s not in interfaces, ignoring plug",
                             interface)
            return

        # ip link set dev localport master vrf_interface
        with self.ip.interfaces[self.vrf_if_idx] as vrf:
            vrf.add_port(interface)

        with self.ip.interfaces[interface] as i:
            i.up()

        # add static ARP entry toward interface (because we can)
        ipr.neigh('add',
                  dst=ip_address,
                  lladdr=mac_address,
                  ifindex=self.ip.interfaces[interface].index,
                  state=netlink.rtnl.ndmsg.states['permanent'])

        # Setup ARP proxying
        proxy_arp(interface, True)

        # Configure gateway address on this port
        # FIXME: that would need to be per vif port
        # Retrieve broadcast IP address
        broadcast_ip = str(netaddr.IPNetwork("%s/%s" % (self.gateway_ip,
                                                        self.network_plen)
                                             ).broadcast)

        try:
            ipr.addr('add', index=self.ip.interfaces[interface].index,
                     address=self.gateway_ip, mask=self.network_plen,
                     broadcast=broadcast_ip)
        except netlink.exceptions.NetlinkError as x:
            if x.code == errno.EEXIST:
                # the route already exists, fine
                self.log.debug("route %s already exists on %s",
                               self.gateway_ip,
                               interface)
            else:
                raise

        # Configure mapping for incoming MPLS traffic
        # with this port label
        req = {'family': pr_common.AF_MPLS,
               'oif': self.ip.interfaces[interface].index,
               'dst': label,  # FIXME how to check for BoS?
               'via': {'family': socket.AF_INET,
                       'addr': ip_address}
               }

        try:
            self.add_route(req)
        except netlink.exceptions.NetlinkError as x:
            if x.code == errno.EEXIST:
                # the route already exists, fine
                self.log.warning("MPLS state for %d already exists", label)
            else:
                raise

    @log_decorator.log_info
    def vif_unplugged(self, mac_address, ip_address, localport, label,
                      direction, last_endpoint=True):
        interface = localport['linuxif']

        if interface not in self.ip.interfaces:
            self.log.warning("interface %s not in interfaces, ignoring plug",
                             interface)
            return

        # bring the interface down, we don't want
        # traffic from this interface to leak out of the VRF
        with self.ip.interfaces[interface] as i:
            i.down()

        # Disable ARP proxying
        proxy_arp(interface, False)

        # ip link set dev localport master vrf_interface
        with self.ip.interfaces[self.vrf_if_idx] as i:
            i.del_port(interface)

        # Unconfigure gateway address on this port
        # FIXME: that would need to be per vif port
        # Retrieve broadcast IP address
        ip = netaddr.IPNetwork("%s/%s" % (self.gateway_ip, self.network_plen))
        broadcast_ip = str(ip.broadcast)

        ipr.addr('del', index=self.ip.interfaces[interface].index,
                 address=self.gateway_ip, mask=self.network_plen,
                 broadcast=broadcast_ip)

        with self.ip.routes.tables['mpls'][label] as r:
            r.remove()

    def _read_mpls_in(self, label):
        routes = [r for r in self.ip.routes.tables["mpls"]
                  if r['dst'] == label]
        assert len(routes) == 1
        res = (routes[0]['oif'], routes[0]['via']['addr'])
        self.log.debug("Found %s for %d with IPDB", res, label)
        return res

    def _nh(self, remote_pe, label, encaps):
        mpls = True
        if str(remote_pe) == self.driver.get_local_address():
            # FIXME: does not work yet, from this table,
            #        'gateway' is considered unreachable
            #        we could drop 'gateway' and just keep oif
            #        but this would only work for connected routes
            # if remote_pe is ourselves
            # we lookup the route for the label and deliver directly
            # to this oif/gateway
            # (oif, gateway) = self._read_mpls_in(label)
            # mpls = False
            gateway = '127.0.0.1'
            oif = self.ip.interfaces['lo'].index
        else:
            gateway = remote_pe
            oif = self.driver.mpls_interface_index

        nh = {'oif': oif}
        if gateway:
            nh['gateway'] = gateway

        if mpls:
            nh['encap'] = {'type': 'mpls',
                           'labels': [{'bos': 1,
                                       'label': label}]}
        self.log.debug("nh: %s", nh)
        return nh

    def _get_route(self, prefix):
        return self.ip.routes.tables[self.rt_table][prefix]

    @log_decorator.log_info
    def setup_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                            nlri, encaps,
                                            lb_consistent_hash_order=0):
        prefix = str(prefix)

        if prefix == "0.0.0.0/0":
            prefix = 'default'

        try:
            with self._get_route(prefix) as r:
                self.log.debug("a route to %s already exists, adding nexthop",
                               prefix)
                r.add_nh(self._nh(remote_pe, label, encaps))
        except KeyError:
            self.log.debug("no route to %s yet, creating", prefix)
            req = {'table': self.rt_table,
                   'dst': prefix,
                   'multipath': [self._nh(remote_pe, label, encaps)]}
            self.log.debug("adding route: %s", req)
            self.add_route(req)

    @log_decorator.log_info
    def remove_dataplane_for_remote_endpoint(self, prefix, remote_pe, label,
                                             nlri, encaps,
                                             lb_consistent_hash_order=0):
        prefix = str(prefix)

        if prefix == "0.0.0.0/0":
            prefix = 'default'

        try:
            with self._get_route(prefix) as r:
                # FIXME: encap info is missing here
                if r['multipath']:
                    r.del_nh(self._nh(remote_pe, label, None))
                else:  # last route
                    r.remove()

        except KeyError:
            self.log.warning("no route found on remove_dataplane_for"
                             "_remote_endpoint for %s", prefix)

    # Looking Glass ##

    def get_lg_map(self):
        return {"routes": (lg.SUBTREE, self.get_lg_routes),
                "route_table": (lg.VALUE, self.rt_table),
                "vrf_if": (lg.VALUE, self.vrf_if),
                }

    @log_decorator.log_info
    def get_lg_routes(self, path_prefix):
        routes = self.ip.routes.tables[self.rt_table]
        return [{r['dst']:
                 jsonutils.loads(jsonutils.dumps(r, default=json_set_default))}
                for r in routes]


class MPLSLinuxDataplaneDriver(dp_drivers.DataplaneDriver):

    """Dataplane driver relying on the MPLS stack in the Linux kernel

    This dataplane driver relies on the MPLS stack in the Linux kernel,
    and on linux vrf interfaces.
    """

    required_kernel = "4.4"
    dataplane_instance_class = MPLSLinuxVRFDataplane
    type = consts.IPVPN
    ecmp_support = True

    driver_opts = [
        cfg.StrOpt("mpls_interface",
                   help=("Interface used to send/receive MPLS traffic. "
                         "Use '*gre*' to choose automatic creation of a tunnel"
                         " interface for MPLS/GRE encap")),
    ]

    def __init__(self):
        super(MPLSLinuxDataplaneDriver, self).__init__()

        privileged_utils.modprobe('mpls_router')
        privileged_utils.modprobe('mpls_gso')
        privileged_utils.modprobe('mpls_iptunnel')
        privileged_utils.modprobe('vrf')

        self.ip = pyroute2.IPDB()  # pylint: disable=no-member

    @log_decorator.log_info
    def reset_state(self):
        # remove all VRF interfaces
        for itf in self.ip.interfaces.keys():
            if isinstance(itf, str) and itf.startswith(VRF_INTERFACE_PREFIX):
                # bring the interface vrf slave interfaces down,
                # we don't want traffic from these interfaces
                # to leak out of the VRF after removal of VRF interface
                for index in self.ip.interfaces[itf].ports:
                    with self.ip.interfaces[index] as port:
                        port.down()

                ipr.link('del', ifname=itf)
        # Flush all routes setup by us in past runs
        ipr.flush_routes(proto=RT_PROT_BAGPIPE)
        # Flush all MPLS routes redirecting traffic to network namespaces
        # (just in case, should be covered by the above)
        ipr.flush_routes(family=pr_common.AF_MPLS)

    @log_decorator.log_info
    def initialize(self):
        sysctl('net.mpls.platform_labels', 2 ** 20 - 1)

        if "*gre*" in self.config["mpls_interface"]:
            self.mpls_interface = "gre_wildcard"
            raise Exception("MPLS/GRE not supported yet")
        else:
            self.mpls_interface = self.config["mpls_interface"]

        sysctl('net.mpls.conf.%s.input' % self.mpls_interface, 1)

        self.mpls_interface_index = self.ip.interfaces[self.mpls_interface
                                                       ].index
        # for traffic from ourselves:
        sysctl('net.mpls.conf.lo.input', 1)

        # enable forwarding
        sysctl('net.ipv4.ip_forward', 1)

    def supported_encaps(self):
        yield exa.Encapsulation(exa.Encapsulation.Type.MPLS)
        # we also accept route with no encap specified
        yield exa.Encapsulation(exa.Encapsulation.Type.DEFAULT)

    # Looking glass ####

    def get_lg_map(self):
        return {"mpls": (lg.SUBTREE, self.get_lg_mpls_routes),
                }

    def get_lg_mpls_routes(self, path_prefix):
        return [{r['dst']:
                 jsonutils.loads(jsonutils.dumps(r, default=json_set_default))}
                for r in self.ip.routes.tables['mpls']]



# Copyright (c) 2016 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import fixtures
from oslo_config import cfg
import random

from networking_bagpipe.tests.fullstack.resources.bagpipe_ml2 \
    import config as bagpipe_ml2_cfg
from networking_bagpipe.tests.fullstack.resources.bgpvpn \
    import config as bgpvpn_cfg
from networking_bagpipe.tests.fullstack.resources.common \
    import config
from networking_bagpipe.tests.fullstack.resources.common \
    import config as common_cfg
from networking_bagpipe.tests.fullstack.resources.common \
    import process as common_proc

from neutron_lib import constants

from neutron.agent.linux import ip_lib
from neutron.agent.linux import utils as a_utils
from neutron.common import utils
from neutron.tests.common.exclusive_resources import ip_network
from neutron.tests.common import net_helpers
from neutron.tests.fullstack.resources import environment as neutron_env
from neutron.tests.fullstack.resources import process as neutron_proc


class BaGPipeEnvironmentDescription(neutron_env.EnvironmentDescription):

    def __init__(self, bagpipe_ml2=False, evpn_driver='linux', bgpvpn=False,
                 ipvpn_driver='ovs', ipvpn_encap='gre',
                 network_type='vxlan', mech_drivers='openvswitch,linuxbridge',
                 service_plugins=None):
        super(BaGPipeEnvironmentDescription, self).__init__(
            network_type=network_type,
            l2_pop=not bagpipe_ml2 and bgpvpn,
            mech_drivers=mech_drivers,
            service_plugins=service_plugins,
            arp_responder=not bagpipe_ml2 and bgpvpn
        )

        self.bagpipe_ml2 = bagpipe_ml2
        self.bgpvpn = bgpvpn
        self.evpn_driver = evpn_driver
        self.ipvpn_driver = ipvpn_driver
        self.ipvpn_encap = ipvpn_encap


class BaGPipeHost(neutron_env.Host):

    def __init__(self, env_desc, host_desc,
                 test_name, neutron_config,
                 central_data_bridge, central_external_bridge,
                 bgp_peer, bgp_port):
        super(BaGPipeHost, self).__init__(env_desc, host_desc,
                                          test_name, neutron_config,
                                          central_data_bridge,
                                          central_external_bridge)
        self.bgp_peer = bgp_peer
        self.bgp_port = bgp_port

    def _setUp(self):
        if (self.env_desc.bgpvpn and
                self.host_desc.l2_agent_type == constants.AGENT_TYPE_OVS):
            self.mpls_bridge = self.useFixture(
                net_helpers.OVSBridgeFixture(self.generate_mpls_bridge())
            ).bridge
            self.mpls_bridge.set_secure_mode()

        super(BaGPipeHost, self)._setUp()

        self.setup_host_with_bagpipe_bgp()

    def generate_mpls_bridge(self):
        return utils.get_rand_device_name(prefix='br-mpls')

    def setup_host_with_ovs_agent(self):
        agent_cfg_fixture = config.OVSConfigFixture(
            self.env_desc, self.host_desc, self.neutron_config.temp_dir,
            self.local_ip, self.mpls_bridge.br_name)
        self.useFixture(agent_cfg_fixture)

        self.useFixture(
            net_helpers.OVSBridgeFixture(
                agent_cfg_fixture.get_br_tun_name())).bridge

        self.ovs_agent = self.useFixture(
            neutron_proc.OVSAgentFixture(
                self.env_desc, self.host_desc,
                self.test_name, self.neutron_config, agent_cfg_fixture))

    def setup_host_with_linuxbridge_agent(self):
        self.host_namespace = self.useFixture(
            net_helpers.NamespaceFixture(prefix="host-")
        ).name

        self.connect_namespace_to_control_network()

        agent_cfg_fixture = config.LinuxBridgeConfigFixture(
            self.env_desc, self.host_desc,
            self.neutron_config.temp_dir,
            self.local_ip,
            physical_device_name=self.host_port.name
        )
        self.useFixture(agent_cfg_fixture)

        agent_fixture_cls = neutron_proc.LinuxBridgeAgentFixture

        self.linuxbridge_agent = self.useFixture(
            agent_fixture_cls(
                self.env_desc, self.host_desc,
                self.test_name, self.neutron_config, agent_cfg_fixture,
                namespace=self.host_namespace
            )
        )

    def setup_host_with_bagpipe_bgp(self):
        if self.host_desc.l2_agent_type == constants.AGENT_TYPE_OVS:
            mpls_bridge = (self.mpls_bridge.br_name
                           if self.env_desc.bgpvpn
                           else '')
            mpls_interface = ''
            if self.env_desc.bgpvpn:
                self.connect_to_internal_network_via_tunneling()

                if self.env_desc.ipvpn_encap == 'bare-mpls':
                    self.connect_to_internal_network_via_mpls_bridge()
                    mpls_interface = filter(
                        lambda port: net_helpers.VETH0_PREFIX
                        in port, self.mpls_bridge.get_port_name_list())[0]
        elif self.host_desc.l2_agent_type == constants.AGENT_TYPE_LINUXBRIDGE:
            mpls_bridge = None
            mpls_interface = self.host_port.name

        bgp_cfg_fixture = common_cfg.BagpipeBGPConfigFixture(
            self.env_desc, self.host_desc, self.neutron_config.temp_dir,
            self.local_ip, self.bgp_peer, self.bgp_port,
            mpls_bridge, mpls_interface)
        self.useFixture(bgp_cfg_fixture)

        self.bagpipe_bgp = self.useFixture(
            common_proc.BagpipeBGPFixture(
                self.env_desc, self.host_desc,
                self.test_name, bgp_cfg_fixture,
                namespace=self.host_namespace))

    def connect_to_internal_network_via_mpls_bridge(self):
        veth_1, veth_2 = self.useFixture(
            net_helpers.VethFixture()).ports

        veth_1.link.set_up()
        veth_2.link.set_up()

        self.mpls_bridge.add_port(veth_1.name)
        self.central_data_bridge.add_port(veth_2.name)

        mpls_device = ip_lib.IPDevice(self.mpls_bridge.br_name)
        mpls_device.addr.add(utils.ip_to_cidr(self.local_ip, 24))
        mpls_device.link.set_up()

        self.mpls_bridge.remove_all_flows()
        self.mpls_bridge.add_flow(in_port='LOCAL', actions='output:1')
        self.mpls_bridge.add_flow(in_port='1', actions='output:LOCAL')


class GoBGPHost(neutron_env.Host):

    def __init__(self, bgp_address, *args, **kwargs):
        super(GoBGPHost, self).__init__(*args, **kwargs)

        self.bgp_address = bgp_address

        # the right thing would be to use a fixture, but we can't in
        # __init__, so let's gamble :)
        self.bgp_port = random.randint(10000, 60000)

    def _setUp(self):
        gobgp_cfg_fixture = self.useFixture(
            common_cfg.GoBGPConfigFixture(
                self.env_desc, "gobgp",
                self.useFixture(fixtures.TempDir()).path,
                self.bgp_address,
                self.bgp_port,
                [host.local_ip for host in self.hosts]))

        self.useFixture(
            common_proc.GoBGPFixture(self.env_desc, None,
                                     self.test_name, gobgp_cfg_fixture))


class BaGPipeEnvironment(neutron_env.Environment):

    def _bagpipe_host_fixture(self, host_desc, bgp_peer, bgp_port):
        temp_dir = self.useFixture(fixtures.TempDir()).path
        neutron_config = bgpvpn_cfg.NeutronConfigFixture(
            self.env_desc, host_desc, temp_dir,
            cfg.CONF.database.connection, self.rabbitmq_environment)
        self.useFixture(neutron_config)

        return self.useFixture(
            BaGPipeHost(self.env_desc,
                        host_desc,
                        self.test_name,
                        neutron_config,
                        self.central_data_bridge,
                        self.central_external_bridge,
                        bgp_peer, bgp_port))

    def _create_gobgp_host(self, bgp_address):
        return GoBGPHost(bgp_address,
                         self.env_desc,
                         "gobgp",
                         self.test_name,
                         None,
                         self.central_data_bridge,
                         self.central_external_bridge)

    def _dont_be_paranoid(self):
        # we will have many br-mplsXXXXX or host-xxx interfaces on the
        # same subnet (the one for self.env_desc.network_range)
        # and we don't want the IP stack to drop the packets received
        # on these because they are from "us" but coming from "the outside"
        a_utils.execute(['sudo', 'sysctl', '-w',
                         'net.ipv4.conf.default.accept_local=1'])
        a_utils.execute(['sudo', 'sysctl', '-w',
                         'net.ipv4.conf.all.rp_filter=0'])
        a_utils.execute(['sudo', 'sysctl', '-w',
                         'net.ipv4.conf.default.rp_filter=0'])

    def _get_network_range(self):
        # for bare MPLS all compute nodes must be in the same subnet
        if self.env_desc.ipvpn_encap == 'bare-mpls':
            self._dont_be_paranoid()
            return self.useFixture(
                ip_network.ExclusiveIPNetwork(
                    "240.0.0.0", "240.255.255.255", "24")).network

        r = super(BaGPipeEnvironment, self)._get_network_range()
        if r:
            self._dont_be_paranoid()
            return r

    def _setUp(self):
        self.temp_dir = self.useFixture(fixtures.TempDir()).path

        # We need this bridge before rabbit and neutron service will start
        self.central_data_bridge = self.useFixture(
            net_helpers.OVSBridgeFixture('cnt-data')).bridge
        self.central_external_bridge = self.useFixture(
            net_helpers.OVSBridgeFixture('cnt-ex')).bridge

        # Get rabbitmq address (and cnt-data network)
        rabbitmq_ip_address = self._configure_port_for_rabbitmq()
        self.rabbitmq_environment = self.useFixture(
            neutron_proc.RabbitmqEnvironmentFixture(host=rabbitmq_ip_address)
        )

        plugin_cfg_fixture = self.useFixture(
            bagpipe_ml2_cfg.ML2ConfigFixture(
                self.env_desc, self.hosts_desc, self.temp_dir,
                self.env_desc.network_type))
        neutron_cfg_fixture = self.useFixture(
            bgpvpn_cfg.NeutronConfigFixture(
                self.env_desc, None, self.temp_dir,
                cfg.CONF.database.connection, self.rabbitmq_environment))

        service_cfg_fixtures = list()
        if self.env_desc.bgpvpn:
            service_cfg_fixtures.append(self.useFixture(
                bgpvpn_cfg.BGPVPNProviderConfigFixture(
                    self.env_desc, self.hosts_desc, self.temp_dir)))

        self.neutron_server = self.useFixture(
            neutron_proc.NeutronServerFixture(
                self.env_desc, None,
                self.test_name, neutron_cfg_fixture, plugin_cfg_fixture,
                service_cfg_fixtures))

        gobgp_host = self._create_gobgp_host(rabbitmq_ip_address)

        self.hosts = [self._bagpipe_host_fixture(desc,
                                                 rabbitmq_ip_address,
                                                 gobgp_host.bgp_port)
                      for desc in self.hosts_desc]

        gobgp_host.hosts = self.hosts

        self.useFixture(gobgp_host)

        self.wait_until_env_is_up()



# Copyright (c) 2017 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from neutron_lib import context as n_context
from neutron_lib.db import api as db_api
from neutron_lib.db import model_base


LOG = logging.getLogger(__name__)

sfc_bagpipe_opts = [
    cfg.IntOpt('as_number', default=64512,
               help=_("Autonomous System number used to generate BGP Route "
                      "Targets that will be used for Port Chain allocations")),
    cfg.ListOpt('rtnn',
                default=[5000, 5999],
                help=_("List containing <rtnn_min>, <rtnn_max> "
                       "defining a range of BGP Route Targets that will "
                       "be used for Port Chain allocations. This range MUST "
                       "not intersect the one used for network segmentation "
                       "identifiers")),
]

cfg.CONF.register_opts(sfc_bagpipe_opts, "sfc_bagpipe")


class BaGPipePpgRTAssoc(model_base.BASEV2, model_base.HasId):
    '''Mapping from Port Pair Group to the NN part of an AS:NN route target'''

    __tablename__ = 'sfc_bagpipe_ppg_rtnn_associations'
    ppg_id = sa.Column(sa.String(36), primary_key=True)
    rtnn = sa.Column(sa.Integer, unique=True, nullable=False)
    is_redirect = sa.Column(sa.Boolean(), nullable=False)
    reverse = sa.Column(sa.Boolean(), nullable=False)


def singleton(class_):
    instances = {}

    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance


@singleton
class RTAllocator(object):
    def __init__(self):
        self.config = cfg.CONF.sfc_bagpipe
        self.ctx = n_context.get_admin_context()
        self.session = self.ctx.session

    def _get_rt_from_rtnn(self, rtnn):
        return ':'.join([str(self.config.as_number), str(rtnn)])

    @log_helpers.log_method_call
    def allocate_rt(self, ppg_id, is_redirect=False, reverse=False):
        ctx = n_context.get_admin_context()
        with db_api.CONTEXT_READER.using(ctx):
            query = ctx.session.query(
                BaGPipePpgRTAssoc).order_by(
                BaGPipePpgRTAssoc.rtnn)

            allocated_rtnns = {obj.rtnn for obj in query.all()}

        # Find first one available in range
        start, end = int(self.config.rtnn[0]), int(self.config.rtnn[1]) + 1
        for rtnn in range(start, end):
            if rtnn not in allocated_rtnns:
                with db_api.CONTEXT_WRITER.using(ctx):
                    ppg_rtnn = BaGPipePpgRTAssoc(ppg_id=ppg_id,
                                                 rtnn=rtnn,
                                                 is_redirect=is_redirect,
                                                 reverse=reverse)
                    ctx.session.add(ppg_rtnn)

                return self._get_rt_from_rtnn(rtnn)

        LOG.error("Can't allocate route target, all range in use")
        return None

    @log_helpers.log_method_call
    def get_rts_by_ppg(self, ppg_id):
        ctx = n_context.get_admin_context()
        with db_api.CONTEXT_READER.using(ctx):
            ppg_rts = ctx.session.query(
                BaGPipePpgRTAssoc).filter_by(ppg_id=ppg_id).all()
            if not ppg_rts:
                return None

        return [ppg_rt.rtnn for ppg_rt in ppg_rts]

    @log_helpers.log_method_call
    def get_redirect_rt_by_ppg(self, ppg_id, reverse=False):
        ctx = n_context.get_admin_context()
        with db_api.CONTEXT_READER.using(ctx):
            ppg_redirect_rt = ctx.session.query(
                BaGPipePpgRTAssoc).filter_by(ppg_id=ppg_id,
                                             is_redirect=True,
                                             reverse=reverse).one()

            return self._get_rt_from_rtnn(ppg_redirect_rt.rtnn)

    @log_helpers.log_method_call
    def release_rt(self, rtnn):
        ctx = n_context.get_admin_context()
        with db_api.CONTEXT_WRITER.using(ctx):
            ppg_rtnn = ctx.session.query(
                BaGPipePpgRTAssoc).filter_by(rtnn=rtnn).first()

            if ppg_rtnn:
                ctx.session.delete(ppg_rtnn)
            else:
                LOG.warning("Can't release route target %s, not used", rtnn)


class BaGPipeChainHop(model_base.BASEV2, model_base.HasId,
                      model_base.HasProject):
    __tablename__ = 'sfc_bagpipe_chain_hops'
    ingress_ppg = sa.Column(sa.String(36),
                            nullable=True)
    egress_ppg = sa.Column(sa.String(36),
                           nullable=True)
    ingress_network = sa.Column(sa.String(36),
                                sa.ForeignKey('networks.id'),
                                nullable=True)
    egress_network = sa.Column(sa.String(36),
                               sa.ForeignKey('networks.id'),
                               nullable=True)
    rts = sa.Column(sa.String(255))
    ingress_gw = sa.Column(sa.String(64))
    egress_gw = sa.Column(sa.String(64))
    readv_from_rts = sa.Column(sa.String(255))
    readv_to_rt = sa.Column(sa.String(255))
    attract_to_rt = sa.Column(sa.String(255))
    redirect_rts = sa.Column(sa.String(255))
    classifiers = sa.Column(sa.String(255))
    reverse_hop = sa.Column(sa.Boolean(), nullable=False)
    portchain_id = sa.Column(sa.String(36),
                             nullable=False,
                             primary_key=True)


class BaGPipeSfcDriverDB(object):

    def initialize(self):
        self.admin_context = n_context.get_admin_context()



#!/usr/bin/env python
#
# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import urllib.error
import urllib.request

import optparse
from oslo_config import cfg
from oslo_serialization import jsonutils

from networking_bagpipe.bagpipe_bgp.api import config as api_config


LOOKING_GLASS_BASE = "looking-glass"


def request(options, server, args):
    target_url = "http://%s:%d/%s/%s" % (server, options.port,
                                         options.prefix, "/".join(args))
    try:
        os.environ['NO_PROXY'] = server
        os.environ['no_proxy'] = server
        response = urllib.request.urlopen(target_url)

        if response.getcode() == 200:
            return jsonutils.load(response)
    except urllib.error.HTTPError as e:
        if e.code == 404:
            return {}
        print("Error requesting %s" % target_url)
        raise
    except Exception:
        print("Error requesting %s" % target_url)
        raise


def normalize(string):
    return (string
            .replace("target:", "rt:")
            .replace(":", "_")
            .replace("-", "_")
            .replace(".", "_"))


RT_TXT_STYLE = 'color="orange",fontcolor="orange"'
RT_STYLE = 'color="orange",fontcolor="orange",arrowhead=onormal'
RT_STYLE_REDIR_FROM_RT = ('color="orange",fontcolor="orange",style=dashed,'
                          'arrowhead=diamond,arrowtail=oinv,headclip=false,'
                          'dir=both')
RT_STYLE_REDIR_TO_RT = ('color="orange",fontcolor="orange",style=dashed,'
                        'arrowhead=onormal,tailclip=false,dir=both,'
                        'arrowtail=oinv')
RT_STYLE_FLOWSPEC = 'color="red",fontcolor="red",style=dotted'
RT_STYLE_FLOWSPEC_INTER = ('color="red",fontcolor="red",style=dotted,'
                           'arrowhead=none,dir=both,arrowtail=oinv')
RT_STYLE_FLOWSPEC_ACTION = 'color="red",fontcolor="red",style=dashed'
PORT_LINK_STYLE = 'style=dashed,dir=none,color="gray",weight=3'


def get_all(options):
    rts = set()
    vpns = set()

    for server in options.servers:
        for vpn in request(options, server, ["vpns", "instances"]):
            vpn_id = vpn['id']
            type = vpn['name'][:3]
            vpns.add((server, vpn_id, type))
            for rt_i in request(options, server, ["vpns", "instances", vpn_id,
                                                  "route_targets", "import"]):
                rts.add(normalize(rt_i))
            for rt_e in request(options, server, ["vpns", "instances", vpn_id,
                                                  "route_targets", "export"]):
                rts.add(normalize(rt_e))

            readvertise = request(options, server, ["vpns", "instances",
                                                    vpn_id, "readvertise"])
            if readvertise:
                rts.add(normalize(readvertise['to'][0]))
                rts.add(normalize(readvertise['from'][0]))

                attract = readvertise.get('attract_traffic', None)
                if attract:
                    rts |= set([normalize(rt)
                                for rt in attract['redirect_rts']])

    return (rts, vpns)


def vpn_uid(server, vpn):
    return "vpn_%s__%s" % (normalize(server), normalize(vpn))


def vpn_short(vpn):
    if len(vpn) > 11:
        return vpn[:4] + ".." + vpn[-5:]
    else:
        return vpn


def main():
    api_config.register_config()
    try:
        cfg.CONF(args=[],
                 project='bagpipe-impex2dot',
                 default_config_files=['/etc/bagpipe-bgp/bgp.conf'])
        api_port = cfg.CONF.API.port
    except cfg.ConfigFilesNotFoundError:
        api_port = api_config.DEFAULT_PORT

    usage = """ %prog [options]

Example: bagpipe-impex2dot --server s1 --server s2 | dot -Tpdf > impex.pdf
    """
    parser = optparse.OptionParser(usage)

    parser.add_option(
        "--server", dest="servers", default=[], action="append",
        help="IP address of a BaGPipe BGP instances (default: localhost)")

    parser.add_option(
        "--port", dest="port", type="int", default=api_port,
        help="Port of BaGPipe BGP (optional, default: %default)")

    parser.add_option(
        "--prefix", dest="prefix", default=LOOKING_GLASS_BASE,
        help="Looking-glass URL Prefix (optional, default: %default)")

    (options, _) = parser.parse_args()

    if len(options.servers) == 0:
        options.servers = ["localhost"]

    ports = set()
    dests = set()
    print('digraph import_export {')

    print('   node [fontname="Helvetica"];')
    print('   edge [fontname="Helvetica"];')
    print('   nodesep=0.55;')

    (rts, vpns) = get_all(options)
    print('   subgraph rts {')
    print('       rank=same;')
    for rt in rts:
        label = rt.upper().replace('_', '\\n')
        print('        %s [shape="circle",label="%s",%s];' % (rt, label,
                                                              RT_TXT_STYLE))
    print('    }')

    print('    subgraph ipvpns {')
    print('        rank=same;')
    for (server, vpn, _) in filter(lambda x: x[2] == 'VRF', vpns):
        server_spec = ("\\n[%s]" % server) if server != 'localhost' else ''
        print('        %s [label="{<0>VRF\\n%s%s|<readv>}",'
              'shape="record"];' % (vpn_uid(server, vpn), vpn_short(vpn),
                                    server_spec))
    print('    }')

    print('    subgraph evpns {')
    for (server, vpn, _) in filter(lambda x: x[2] == 'EVI', vpns):
        server_spec = ("\\n[%s]" % server) if server != 'localhost' else ''
        print('        %s [label="{<0>EVI\\n%s%s}",'
              'shape="record"];' % (vpn_uid(server, vpn), vpn_short(vpn),
                                    server_spec))
    print('    }')

    for (server, vpn, _) in vpns:
        print('    /* %s:%s */' % (server, vpn))
        uid = vpn_uid(server, vpn)

        for rt_i in request(options, server, ["vpns", "instances", vpn,
                                              "route_targets", "import"]):
            print('    %s -> %s:0 [%s];' % (normalize(rt_i), uid,
                                            RT_STYLE))

        for rt_e in request(options, server, ["vpns", "instances", vpn,
                                              "route_targets", "export"]):
            print('    %s:0 -> %s [%s];' % (uid, normalize(rt_e),
                                            RT_STYLE))

        readvertise = request(options, server, ["vpns", "instances",
                                                vpn, "readvertise"])
        if readvertise:
            readv = "%s:readv" % vpn_uid(server, vpn)
            print('    %s -> %s [label="",%s];' %
                  (normalize(readvertise['from'][0]),
                   readv,
                   RT_STYLE_REDIR_FROM_RT))
            print('    %s -> %s [label="",%s];' %
                  (readv,
                   normalize(readvertise['to'][0]),
                   RT_STYLE_REDIR_TO_RT))

            attract = readvertise.get('attract_traffic', None)
            if attract:
                intermediate = "%s_%s_attract" % (normalize(server),
                                                  normalize(vpn))
                print('    %s [style=invis,height=0,width=0,'
                      'fixedsize=true,rank=1]' % intermediate)

                redir_rt = attract['redirect_rts'][0]

                # readv -> intermediate
                print('    %s -> %s [label="flow",%s];' %
                      (readv, intermediate,
                       RT_STYLE_FLOWSPEC_INTER))

                # intermediate -> flowspec route RT
                print('    %s -> %s [label="",weight=50,%s];' %
                      (intermediate, normalize(redir_rt),
                       RT_STYLE_FLOWSPEC))

                # intermediate -> flowspec route redirect action  RT
                print('    %s -> %s [label="action",%s,arrowhead=none'
                      ',weight=20];' %
                      (intermediate, normalize(readvertise['to'][0]),
                       RT_STYLE_FLOWSPEC_ACTION))

        for port in request(options, server, ["vpns", "instances",
                                              vpn, "ports"]).keys():
            print('    %s -> port_%s_%s [%s,weight=5];' % (uid,
                                                           normalize(server),
                                                           normalize(port),
                                                           PORT_LINK_STYLE))
            ports.add((server, normalize(port)))

        # possible link between an E-VPN and an IPVPN ?
        ipvpn = request(options, server, ["vpns", "instances",
                                          vpn, "gateway_port", "ipvpn"])
        if ipvpn:
            ipvpn_id = ipvpn['external_instance_id']
            print('    %s -> %s [weight=500];' % (uid,
                                                  vpn_uid(server, ipvpn_id))
                  )

    for (server, port) in ports:
        print('    port_%s_%s [label="",style=invis,height=0,width=0,'
              'fixedsize=true];' % (normalize(server),
                                    normalize(port)))
        if port.startswith('to_'):
            dest = port.split('_')[1]
            print('    port_%s_%s -> dest_%s_%s [style=dashed,'
                  'dir=none,color="gray"'
                  ',weight=5];' %
                  (normalize(server), normalize(port), normalize(server),
                   dest))
            dests.add((server, dest))

    for (server, dest) in dests:
        print('    dest_%s_%s [label="%s\\n[%s]",shape="square",'
              'color="gray"];' % (normalize(server),
                                  normalize(dest), dest, server))

    print('}')



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import socket

import httplib2
import json

from oslo_config import cfg

from oslo_log import helpers as log_helpers
from oslo_log import log as logging

from oslo_concurrency import lockutils
from oslo_serialization import jsonutils
from oslo_service import loopingcall

from networking_bagpipe._i18n import _

from networking_bagpipe.bagpipe_bgp import constants as bbgp_const

from neutron.conf.agent import common as config

from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc

LOG = logging.getLogger(__name__)


# Having this at line 231 is apparently not enough, so adding here as well:
# pylint: disable=not-callable

bagpipe_bgp_opts = [
    cfg.IntOpt('ping_interval', default=10,
               help=_("The number of seconds the bagpipe-bgp client will "
                      "wait between polling for restart detection.")),
    cfg.PortOpt('bagpipe_bgp_port', default=8082,
                help=_("bagpipe-bgp REST service IP port.")),
]

# these options are for internal use only (fullstack tests), and hence
# better kept in a separate table not looked at by oslo gen confi hooks
internal_opts = [
    cfg.HostAddressOpt('bagpipe_bgp_ip', default='127.0.0.1',
                       help=_("bagpipe-bgp REST service IP address.")),
]

cfg.CONF.register_opts(bagpipe_bgp_opts, "BAGPIPE")
cfg.CONF.register_opts(internal_opts, "BAGPIPE")
config.register_agent_state_opts_helper(cfg.CONF)

# don't use bbgp_const.VPN_TYPES here, because here in this module
# we sometimes need to iterate the vpn types in a controlled order:
# EVPN first on attach, EVPN last on detach
VPN_TYPES = [bbgp_const.EVPN, bbgp_const.IPVPN]


class BaGPipeBGPException(n_exc.NeutronException):
    message = "An exception occurred when calling bagpipe-bgp \
               REST service: %(reason)s"


class SetJSONEncoder(jsonutils.JSONEncoder):
    # JSON encoder that encodes set like a list, this
    # allows to store list of RTs as sets and simplify the code
    # in many places
    # pylint: disable=method-hidden
    def default(self, obj):
        if isinstance(obj, set):
            return list(obj)
        return json.JSONEncoder.default(self, obj)


class HTTPClientBase(object):
    """An HTTP client base class"""

    def __init__(self, host="127.0.0.1", port=8082,
                 client_name="HTTP client base"):
        """Create a new HTTP client

        :param host: HTTP server IP address
        :param port: HTTP server port
        """
        self.host = host
        self.port = port
        self.client_name = client_name

    def do_request(self, method, action, body=None):
        LOG.debug("bagpipe-bgp client request: %(method)s %(action)s "
                  "[%(body)s]",
                  {'method': method, 'action': action, 'body': str(body)})

        if isinstance(body, dict):
            body = jsonutils.dumps(body, cls=SetJSONEncoder)
        try:
            headers = {'User-Agent': self.client_name,
                       "Content-Type": "application/json",
                       "Accept": "application/json"}
            uri = "http://%s:%s/%s" % (self.host, self.port, action)

            http = httplib2.Http()
            response, content = http.request(uri, method, body, headers)
            LOG.debug("bagpipe-bgp returns [%(status)s:%(content)s]",
                      {'status': str(response.status), 'content': content})

            if response.status == 200:
                if content and len(content) > 1:
                    return jsonutils.loads(content)
            else:
                reason = (
                    "An HTTP operation has failed on bagpipe-bgp."
                )
                raise BaGPipeBGPException(reason=reason)
        except (socket.error, IOError) as e:
            reason = "Failed to connect to bagpipe-bgp: %s" % str(e)
            raise BaGPipeBGPException(reason=reason)

    def get(self, action):
        return self.do_request("GET", action)

    def post(self, action, body=None):
        return self.do_request("POST", action, body=body)

    def put(self, action, body=None):
        return self.do_request("PUT", action, body=body)

    def delete(self, action):
        return self.do_request("DELETE", action)


def get_default_vpn_instance_id(vpn_type, network_id):
    return '%s_%s' % (vpn_type, network_id)


class BaGPipeBGPAgent(HTTPClientBase):
    """Implements a BaGPipe-BGP REST client"""

    _instance = None

    # bagpipe-bgp status
    BAGPIPEBGP_UP = 'UP'
    BAGPIPEBGP_DOWN = 'DOWN'

    @classmethod
    @lockutils.synchronized('bagpipe-bgp-agent')
    def _create_instance(cls, agent_type):
        if not cls.has_instance():
            cls._instance = cls(agent_type)

    @classmethod
    def has_instance(cls):
        return cls._instance is not None

    @classmethod
    def clear_instance(cls):
        cls._instance = None

    @classmethod
    def get_instance(cls, agent_type):
        # double checked locking
        if not cls.has_instance():
            cls._create_instance(agent_type)
        else:
            if cls._instance.agent_type != agent_type:
                raise Exception("Agent already configured with another type")

        return cls._instance

    def __init__(self, agent_type):

        """Create a new BaGPipe-BGP REST service client.

        :param agent_type: bagpipe-bgp agent type (Linux bridge or OVS)

        """
        super(BaGPipeBGPAgent,
              self).__init__(cfg.CONF.BAGPIPE.bagpipe_bgp_ip,
                             cfg.CONF.BAGPIPE.bagpipe_bgp_port, agent_type)

        self.agent_type = agent_type

        self.ping_interval = cfg.CONF.BAGPIPE.ping_interval

        self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN
        self.seq_num = 0

        # Starts a greenthread for bagpipe-bgp status polling
        self._start_bagpipe_bgp_status_polling(self.ping_interval)

        # Maps registered service name to callback function used to build the
        # content of bagpipe-bgp API calls
        self.build_callbacks = dict()

        # Maps registered service name to list of ports for which a given
        # service is enabled
        self.port_lists = dict()

    def _check_bagpipe_bgp_status(self):
        """Trigger refresh on bagpipe-bgp restarts

        Check if bagpipe-bgp has restarted while sending ping request
        to detect sequence number change.
        If a restart is detected, re-send all registered attachments to
        bagpipe-bgp.
        """
        new_seq_num = self._request_ping()

        # Check bagpipe-bgp restart
        if new_seq_num != self.seq_num:
            if new_seq_num != -1:
                if self.seq_num != 0:
                    LOG.warning("bagpipe-bgp restart detected...")
                else:
                    LOG.info("bagpipe-bgp successfully detected")

                self.seq_num = new_seq_num
                self.bagpipe_bgp_status = self.BAGPIPEBGP_UP

                # Re-send all registered ports to bagpipe-bgp
                ports = set()
                for port_list in self.port_lists.values():
                    ports |= port_list

                if ports:
                    LOG.info("Sending all registered ports to bagpipe-bgp")
                    LOG.debug("Registered ports list: %s", ports)
                    for port_id in ports:
                        self.do_port_plug(port_id)
                else:
                    LOG.info("No attachment to send to bagpipe-bgp")
            else:
                self.bagpipe_bgp_status = self.BAGPIPEBGP_DOWN

    def _start_bagpipe_bgp_status_polling(self, ping_interval=10):
        # Start bagpipe-bgp status polling at regular interval
        status_loop = loopingcall.FixedIntervalLoopingCall(
            self._check_bagpipe_bgp_status)
        status_loop.start(interval=ping_interval,
                          initial_delay=ping_interval)  # TM: why not zero ?

    def _check_evpn2ipvpn_info(self, vpn_type, network_id, attach_list,
                               attach_info):
        # Check if plugging an EVPN into an IPVPN
        if (vpn_type == bbgp_const.IPVPN and bbgp_const.EVPN in attach_list):
            attach_info['local_port'] = {
                bbgp_const.EVPN: {
                    'id': get_default_vpn_instance_id('evpn', network_id)
                }
            }

    def _compile_port_attach_info(self, port_id):
        # this method returns information for all bagpipe-bgp attachments to
        # produce for a given port:
        # {
        #   'evpn': [
        #       {
        #          'vpn_instance_id':
        #          'ip_address'
        #          ..
        #          'import_rts': [...]
        #          'export_rts': [...]
        #       },
        #       ...
        #   ],
        #   'ipvpn': [
        #       {
        #          'vpn_instance_id':
        #          'ip_address'
        #          ..
        #          'import_rts': [...]
        #          'export_rts': [...]
        #       },
        #       ...
        #   ]
        # }
        #
        # This structure produces consolidated information across all services
        # for all Route Target parameters: the import_rt and export_rt
        # attributes accumulate the RTs of all the service producing
        # attachments for a given VPN instance (vpn instance id).
        #
        # Another consolidation that is done in the case where both EVPN and
        # IPVPN attachments are produced, and the agent is a linuxbridge agent.
        # In that case the IPVPN attachments are modified so that instead of
        # plugging the port into the IPVPN, the EVPN instance is plugged into
        # the IPVPN, which is necessary so that bagpipe-bgp will connect the
        # linux bridge to a linux VRF with a veth interface.
        #
        # NOTE(tmorin): the code does not do consistency checks for parameters
        # that would be conflicting between attachments, for instance if
        # an EVPN attachment would specify a VNI X and another a VNI Y.
        # For read-only parameters, bagpipe-bgp would be in charge of detecting
        # an attempt at overwriting the parameters with a different value and
        # raising an error.
        # For proper safeguarding, the code here would need to check
        # consistency across attachments of read-write values.

        service_attachments_map = {}
        for service, build_callback in self.build_callbacks.items():
            # what was returned by the callbacks before (a dict allowing to
            # describe one EVPN and one IPVPN attachment):
            # {
            #    'network_id':
            #    'ip_address'
            #    ..
            #    'evpn': {
            #         'import_rt': [...]
            #         'export_rt': [...]
            #         'static_routes': ...
            #    }
            #    'ipvpn': {
            #         'import_rt': [...]
            #         'export_rt': [...]
            #         'static_routes': ...
            #    }
            # }
            #
            # we expect the callback to return a dict or lists,
            # following this template:
            # {
            #    'network_id': # use to generate vpn_instance_id if
            #                  # vpn_instance_id is omitted below
            #    'evpn': [
            #       {
            #         'vpn_instance_id': ..
            #         'ip_address': ..
            #         ...
            #         'import_rt': [...]
            #         'export_rt': [...]
            #       },
            #       ...
            #    ],
            #    'ipvpn': [
            #    ]
            # }
            service_attachments_map[service] = build_callback(port_id)
            LOG.debug("port %s, attach info for %s: %s",
                      port_id, service, service_attachments_map[service])

        attach_list = {}

        # map in which we consolidate the RTs for a given vpn instance
        # vpn_instance_rts[vpn_instance_id]['import_rt'] = set()
        # vpn_instance_rts[vpn_instance_id]['export_rt'] = set()
        vpn_instance_rts = {}

        for vpn_type in VPN_TYPES:
            vpn_attachment_list = []

            for service in self.build_callbacks.keys():
                service_attachments = (
                    service_attachments_map[service].get(vpn_type)
                )

                if not service_attachments:
                    continue

                default_vpn_instance_id = get_default_vpn_instance_id(
                    vpn_type,
                    service_attachments_map[service]['network_id'])

                for service_attachment in service_attachments:

                    vpn_instance_id = service_attachment.setdefault(
                        'vpn_instance_id', default_vpn_instance_id)

                    service_attachment['vpn_type'] = vpn_type

                    # initialize consolidated RTs for this vpn_instance_id
                    # if this wasn't done yet
                    vpn_instance_rts.setdefault(vpn_instance_id, {
                        bbgp_const.RT_IMPORT: set(),
                        bbgp_const.RT_EXPORT: set()
                    })

                    for rt_type in (bbgp_const.RT_IMPORT,
                                    bbgp_const.RT_EXPORT):
                        # merge this service RTs with the RTs we already had
                        # for this vpn_instance_id
                        orig_rts = set(service_attachment[rt_type])
                        vpn_instance_rts[vpn_instance_id][rt_type] |= orig_rts
                        # have the RT information for this attachment
                        # point to the consolidated RT list
                        service_attachment[rt_type] = (
                            vpn_instance_rts[vpn_instance_id][rt_type])

                    LOG.debug("adding processed attachment: %s",
                              service_attachment)
                    vpn_attachment_list.append(service_attachment)

            if vpn_attachment_list:
                attach_list[vpn_type] = vpn_attachment_list

        if self.agent_type == n_const.AGENT_TYPE_LINUXBRIDGE:
            if (attach_list.get(bbgp_const.EVPN) and
                    attach_list.get(bbgp_const.IPVPN)):
                # go through all IPVPN attachments and rewrite local_port
                # to point to the evpn instance, rather than the VM port
                for attachment in attach_list[bbgp_const.IPVPN]:
                    attachment['local_port'] = {bbgp_const.EVPN: {
                        'id': attachment['vpn_instance_id'].replace('ipvpn',
                                                                    'evpn')
                    }}

        LOG.debug("all attachments for port %s: %s", port_id, attach_list)
        return attach_list

    def _request_ping(self):
        """Send ping request to bagpipe-bgp to get sequence number"""
        try:
            response = self.get('ping')
            LOG.debug("bagpipe-bgp PING response received with "
                      "sequence number %s", response)
            return response
        except Exception as e:
            LOG.warning(str(e))
            return -1

    @log_helpers.log_method_call
    def _send_attach_local_port(self, local_port_details):
        """Send local port attach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('attach_localport', local_port_details)
                LOG.debug("Local port has been attached to bagpipe-bgp with "
                          "details %s", local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't attach local port on bagpipe-bgp: %s", str(e))
        else:
            LOG.debug("Local port not yet attached to bagpipe-bgp (not up)")

    @log_helpers.log_method_call
    def _send_detach_local_port(self, local_port_details):
        """Send local port detach request to BaGPipe-BGP if running"""
        if self.bagpipe_bgp_status is self.BAGPIPEBGP_UP:
            try:
                self.post('detach_localport', local_port_details)
                LOG.debug("Local port has been detached from bagpipe-bgp "
                          "with details %s", local_port_details)
            except BaGPipeBGPException as e:
                LOG.error("Can't detach local port from bagpipe-bgp: %s",
                          str(e))
                raise
        else:
            LOG.debug("Local port not yet detached from bagpipe-bgp (not up)")

    @log_helpers.log_method_call
    def _send_all_attachments(self, plug_details):
        # First plug E-VPNs because they could be plugged into IP-VPNs
        for vpn_type in [t for t in VPN_TYPES if t in plug_details]:
            for plug_detail in plug_details[vpn_type]:
                self._send_attach_local_port(plug_detail)

    @log_helpers.log_method_call
    def do_port_plug(self, port_id):
        """Send port attach request to bagpipe-bgp."""
        all_plug_details = self._compile_port_attach_info(port_id)

        self._send_all_attachments(all_plug_details)

    @log_helpers.log_method_call
    def do_port_plug_refresh(self, port_id, detach_infos):
        """Refresh port attach on bagpipe-bgp

        Send port attach and/or detach request to bagpipe-bgp when necessary.

        detach_infos:
        {
            'network_id': ...
            'evpn': {
                ...
            }
            'ipvpn': {
                ...
            }
        }
        ]
        """
        self.do_port_plug_refresh_many(port_id, [detach_infos])

    @log_helpers.log_method_call
    def do_port_plug_refresh_many(self, port_id, detach_info_list):
        """Refresh port attach on bagpipe-bgp

        Send port attach and/or detach request to bagpipe-bgp when necessary.

        detach_infos, a list of:
        {
            'network_id': ...
            'evpn': {
                ...
            }
            'ipvpn': {
                ...
            }
        }
        ]
        """
        plug_details = self._compile_port_attach_info(port_id)

        for detach_infos in detach_info_list:
            network_id = detach_infos.pop('network_id')
            for detach_vpn_type, detach_info in list(detach_infos.items()):
                detach_info.setdefault(
                    'vpn_instance_id',
                    get_default_vpn_instance_id(detach_vpn_type,
                                                network_id))
                detach_info['vpn_type'] = detach_vpn_type

                # NOTE(tmorin): to be reconsidered
                self._check_evpn2ipvpn_info(detach_vpn_type, network_id,
                                            plug_details, detach_info)

            if detach_infos:
                # unplug IPVPN first, then EVPN (hence ::-1 below)
                for vpn_type in [t for t in VPN_TYPES[::-1]
                                 if t in detach_infos]:
                    self._send_detach_local_port(detach_infos[vpn_type])

        self._send_all_attachments(plug_details)

    @lockutils.synchronized('bagpipe-bgp-agent')
    def register_build_callback(self, service_name, callback):
        self.build_callbacks[service_name] = callback

    @lockutils.synchronized('bagpipe-bgp-agent')
    def register_port_list(self, service_name, port_list):
        self.port_lists[service_name] = port_list



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import itertools

from collections import defaultdict
from collections import namedtuple

from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class keydefaultdict(defaultdict):
    """Inherit defaultdict class to customize default_factory.

    Override __missing__ method to construct custom object as default_factory
    passing an argument.

    For example:
    class C(object):
        def __init__(self, value):
            self.value = value

    d = keydefaultdict(C)
    d[key] returns C(key)
    """
    # pylint: disable=not-callable
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = getattr(self, 'default_factory')(key)
            return ret


class CommonInfo(object):

    def __init__(self, id):
        self.id = id
        self._associations = dict()

    @property
    def associations(self):
        return self._associations.values()

    def add_association(self, association):
        self._associations[association.id] = association

    def remove_association(self, association):
        del self._associations[association.id]

    def has_association(self, association_id):
        return association_id in self._associations

    def get_association(self, association_id):
        return self._associations.get(association_id)


class PortInfo(CommonInfo):

    def __init__(self, port_id):
        super(PortInfo, self).__init__(port_id)

        self.ip_address = None
        self.mac_address = None
        self.network = None
        self.chain_hops = dict()
        self.admin_state_up = False

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and
                self.id == other.id)

    def __hash__(self):
        return hash(self.id)

    @property
    def all_associations(self):
        return itertools.chain(
            self.associations,
            self.network.associations if self.network else [])

    def has_any_association(self):
        return any(True for _ in self.all_associations)

    def add_chain_hop(self, chain_hop):
        if not chain_hop:
            return

        if not all(item in self.chain_hops.items()
                   for item in chain_hop.items()):
            self.chain_hops.update(chain_hop)

    def update_admin_state(self, port_data, transition_to_down_hook=None):
        if port_data['admin_state_up']:
            self.admin_state_up = True
        else:
            previous_admin_state_up = self.admin_state_up
            if previous_admin_state_up:
                if callable(transition_to_down_hook):
                    transition_to_down_hook()
                self.admin_state_up = False
            else:
                self.admin_state_up = False
        LOG.debug("port %s, admin_state_up is now: %s",
                  self.id, self.admin_state_up)

    def __repr__(self):
        return "PortInfo(%s,%s)" % (self.id, self.admin_state_up)


GatewayInfo = namedtuple('GatewayInfo', ['mac', 'ip'])
NO_GW_INFO = GatewayInfo(None, None)


class NetworkInfo(CommonInfo):

    def __init__(self, network_id):
        super(NetworkInfo, self).__init__(network_id)

        self.gateway_info = NO_GW_INFO
        self.ports = set()
        self.segmentation_id = None

    def set_gateway_info(self, gateway_info):
        self.gateway_info = gateway_info

    def __repr__(self):
        return "NetInfo: %s (segmentation id:%s, gw:%s), " % (
            self.id, self.segmentation_id, self.gateway_info)


class BaseInfoManager(object):

    def __init__(self):
        # Store all ports level network and service informations
        self.ports_info = keydefaultdict(PortInfo)
        # Store all networks level network and service informations
        self.networks_info = keydefaultdict(NetworkInfo)

    def _get_network_port_infos(self, net_id, port_id):
        net_info = self.networks_info[net_id]
        port_info = self.ports_info[port_id]
        net_info.ports.add(port_info)
        port_info.network = net_info

        return net_info, port_info

    def _remove_network_port_infos(self, net_id, port_id):
        port_info = self.ports_info.get(port_id)
        net_info = self.networks_info.get(net_id)

        if port_info:
            del self.ports_info[port_id]

        if net_info:
            net_info.ports.discard(port_info)
            if not net_info.ports:
                del self.networks_info[net_id]



# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_tracker_worker
   :synopsis: a module that defines several test cases for the tracker_worker
              module.
   In particular, unit tests for TrackerWorker class.
   Setup: Run TrackerWorker instance.
   TearDown: Stop TrackerWorker instance.
   TrackerWorker is in charge to receive RouteEvent from RouteTableManager.
   A RouteEvent contains an event type ADVERTIZE or WITHDRAW, and a RouteEntry.
   TrackerWorker should call new_best_route and/or best_route_removed if the
   new RouteEntry changes the current list of the known best routes. The
   current list of the known best routes, which can be modified by the new
   RouteEntry, is selected thanks to the tracked_entry associated to the new
   RouteEntry. The tracked_entry is obtained thanks to _route2TrackedEntry.
   _compare_routes is used to compare 2 RouteEntry.
   Unit tests are organized as follow:
   TestA: basic tests, advertise several routes with different NLRI and same or
          different sources
   TestB: same routes (with _compare_routes) announced by different sources
   TestC: different routes (with _compare_routes) announced by different
          sources, TrackerWorker selects the best route.
   TestD: ECMP routes or same routes (with _compare_routes), same source, same
          attributes except NextHop
   TestE: different routes (with compare_routes announced by the same source
          with replaced_route not none
"""

import copy
import threading
from unittest import mock

import testtools

from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import tracker_worker
from networking_bagpipe.bagpipe_bgp.engine import worker
from networking_bagpipe.tests.unit.bagpipe_bgp import base as t


def _test_compare_routes(self, route_a, route_b):
    if (route_a.nlri != route_b.nlri or
            route_a.afi != route_b.afi or
            route_a.safi != route_b.safi):
        raise Exception('Bug: compare_routes called with routes having '
                        'different nlri/afi/safi')
    else:
        if (route_a.attributes.sameValuesAs(route_b.attributes)):
            return 0
        else:
            lp_a = route_a.attributes[exa.Attribute.CODE.LOCAL_PREF].localpref
            nh_a = route_a.attributes[exa.Attribute.CODE.NEXT_HOP].top()

            lp_b = route_b.attributes[exa.Attribute.CODE.LOCAL_PREF].localpref
            nh_b = route_b.attributes[exa.Attribute.CODE.NEXT_HOP].top()

            if nh_a != nh_b and lp_a == lp_b:
                # ECMP routes
                return 0
            else:
                return (lp_a > lp_b) - (lp_b > lp_a)


class TrackerWorkerThread(tracker_worker.TrackerWorker, threading.Thread):

    def __init__(self):
        threading.Thread.__init__(self, name='TrackerWorkerThread')
        self.setDaemon(True)
        tracker_worker.TrackerWorker.__init__(
            self, mock.Mock(), 'TrackerWorker', _test_compare_routes)

    def stop(self):
        self._please_stop.set()
        self._queue.put(worker.STOP_EVENT)
        self._stopped()

    def route_to_tracked_entry(self, route):
        return route.nlri

    # the definitions below are needed because TrackerWorker is an abstract
    # class
    def new_best_route(self, entry, route):
        pass

    def best_route_removed(self, entry, route, last):
        pass


class TestTrackerWorker(testtools.TestCase, t.BaseTestBagPipeBGP):

    def setUp(self):
        super(TestTrackerWorker, self).setUp()
        self.tracker_worker = TrackerWorkerThread()
        self.tracker_worker.start()
        self.set_event_target_worker(self.tracker_worker)
        self._calls = []

    def tearDown(self):
        super(TestTrackerWorker, self).tearDown()
        self.tracker_worker.stop()
        self.tracker_worker.join()

    def _check_calls(self, call_args_list, expected_list, ordered=True):
        # use to check the calls to new_best_route and best_route_removed
        # against a list of expected calls

        expected_list_copy = []
        # clear source field in the routes in expected calls
        # because the new_best_route and best_route_removed do not receive
        # routes with this field set
        for expected in expected_list:
            route = copy.copy(expected[1])
            route.source = None
            self.assertIn(len(expected), (2, 3))
            if len(expected) == 2:
                expected_list_copy.append((expected[0], route))
            elif len(expected) == 3:
                expected_list_copy.append((expected[0], route, expected[2]))

        if not ordered:
            expected_list_copy = sorted(expected_list_copy,
                                        key=repr)
            call_args_list = sorted(call_args_list,
                                    key=lambda x: repr(x[0]))

        for ((call_args, _), expected) in zip(call_args_list,
                                              expected_list_copy):
            self.assertEqual(expected[0], call_args[0], 'Bad prefix')

            observed_route_entry = call_args[1]
            expected_route_entry = expected[1]

            self.assertEqual(expected_route_entry, observed_route_entry)

            if len(expected) >= 3:
                self.assertEqual(expected[2], call_args[2],
                                 "wrong 'last' flag")

    def _call_list(self, method):
        def side_effect(*args, **kwargs):
            self._append_call(method)
        return side_effect

    def test_a1_different_nlri_same_source(self):
        # A source A advertises and withdraws routes for different NLRI.
        # Mock objects
        self.tracker_worker.new_best_route = mock.Mock()
        self.tracker_worker.best_route_removed = mock.Mock()

        # Only 1 source A
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        # Source A advertises a route for NLRI1
        route_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source A advertises a route for NLRI2
        route_nlri2a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI2, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source A withdraws the route for NLRI1
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source A withdraws the route for NLRI2
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI2, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        self.assertEqual(2, self.tracker_worker.new_best_route.call_count,
                         '2 new best routes: 1 for NLRI1 and 1 for NLRI2')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route_nlri1a.route_entry),
                           (t.NLRI2, route_nlri2a.route_entry)])
        self.assertEqual(2, self.tracker_worker.best_route_removed.call_count,
                         '2 old routes removed: 1 for NLRI1 and 1 for NLRI2')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route_nlri1a.route_entry, True),
             (t.NLRI2, route_nlri2a.route_entry, True)])

    def test_a2_different_nlri_different_source(self):
        # 2 sources A and B advertise and withdraw routes for different NLRI.
        # Mock objects
        self.tracker_worker.new_best_route = mock.Mock()
        self.tracker_worker.best_route_removed = mock.Mock()

        # 2 sources: A and B
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')
        # Source A advertises a route for NLRI1
        route_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source B advertises a route for NLRI2
        route_nlri2B = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI2, [t.RT1, t.RT2],
            worker_b, t.NH1, 100)
        # Source A withdraws the route for NLRI1
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source B withdraws the route for NLRI2
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI2, [t.RT1, t.RT2],
            worker_b, t.NH1, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        self.assertEqual(2, self.tracker_worker.new_best_route.call_count,
                         '2 new_best_route calls: 1 for NLRI1 and 1 for NLRI2')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route_nlri1a.route_entry),
                           (t.NLRI2, route_nlri2B.route_entry)])
        self.assertEqual(2, self.tracker_worker.best_route_removed.call_count,
                         '2 best_route_removed calls: 1 for NLRI1 and 1 for '
                         'NLRI2')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route_nlri1a.route_entry, True),
             (t.NLRI2, route_nlri2B.route_entry, True)])

    def test_a3_same_nlri_same_source(self):
        # A source A advertises the same route for the same NLRI
        # Mock objects
        self.tracker_worker.new_best_route = mock.Mock()
        self.tracker_worker.best_route_removed = mock.Mock()

        # 1 source: A
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        # Source A advertises a route for NLRI1
        route_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source A advertises the same route for NLRI1
        self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        self.assertEqual(1, self.tracker_worker.new_best_route.call_count,
                         'expected 1 new_best_route call for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route_nlri1a.route_entry),
                           (t.NLRI1, route_nlri1a.route_entry)])

    def test_a4_withdraw_nlri_not_known(self):
        # A source A withdraws a route that does not exist.
        self.tracker_worker.new_best_route = mock.Mock()
        self.tracker_worker.best_route_removed = mock.Mock()

        # 1 source: A
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        # Source A withdraws a route for NLRI1 which is not known by
        # tracker_worker
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)

        # Check calls to new_best_route and best_route_removed
        self.assertEqual(0, self.tracker_worker.new_best_route.call_count,
                         'new_best_route should not have been called')
        self.assertEqual(0, self.tracker_worker.best_route_removed.call_count,
                         'best_route_removed should not have been called')

    def test_b1_is_the_current_best_route(self):
        # The route which is advertised by another source is the current best
        # route
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 2 sources: A and B
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')

        # Source A advertises a route for NLRI1
        self._append_call("RE1")
        route_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source B advertises the same route for NLRI1
        self._append_call("RE2")
        route_nlri1B = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 100)
        # Source A withdraws the route for NLRI1
        self._append_call("RE3")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source B withdraws the route for NLRI1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        self.assertEqual(
            1, self.tracker_worker.new_best_route.call_count,
            '1 new best route call for NLRI1')
        self._check_calls(
            self.tracker_worker.new_best_route.call_args_list,
            [(t.NLRI1, route_nlri1a.route_entry)])
        self.assertEqual(
            1, self.tracker_worker.best_route_removed.call_count,
            '1 best_route_removed call for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route_nlri1B.route_entry, True)])

        expected_calls = ["RE1", t.NBR, "RE2", "RE3", "RE4", t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

    def test_b2_is_not_the_current_best_route(self):
        # The route which is advertised by an other source is not the current
        # best route but will become the best route
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 3 sources: A, B and C
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')
        worker_c = worker.Worker(mock.Mock(), 'worker.Worker-C')

        # Source A advertises route1 for NLRI1
        self._append_call("RE1")
        route1Nlri1 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B advertises route2 for NLRI1 : route1 is better than route2
        self._append_call("RE2")
        route2Nlri1 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source C advertises also route2
        self._append_call("RE3")
        self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_c, t.NH1, 200)
        # Source A withdraws route1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", "RE3", "RE4", t.NBR, t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new best route call for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1Nlri1.route_entry),
                           (t.NLRI1, route2Nlri1.route_entry)])
        self.assertEqual(
            1, self.tracker_worker.best_route_removed.call_count,
            '1 best_route_removed call for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1Nlri1.route_entry, False)])

    def test_c1_route1_best_route(self):
        # Route1 is the best route
        # Mock objects
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 2 sources : A and B
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B advertises a route2 for NLRI1 with different attributes.
        # Route1 is better than Route2
        self._append_call("RE2")
        route2_nlri1b = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE3")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B withdraws route2 for NLRI1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", "RE3",
                          t.NBR, t.BRR, "RE4", t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route2_nlri1b.route_entry)])
        self.assertEqual(
            2, self.tracker_worker.best_route_removed.call_count,
            '2 best_route_removed calls for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False),
             (t.NLRI1, route2_nlri1b.route_entry, True)])

    def test_c2_route2_best_route(self):
        # Route2 is the best route
        # Mock objects
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 2 sources: A and B
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source B advertises a route2 for NLRI1. Route2 is better than Route1
        self._append_call("RE2")
        route2_nlri1b = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE3")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", t.NBR, t.BRR, "RE3"]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route2_nlri1b.route_entry)])
        self.assertEqual(
            1, self.tracker_worker.best_route_removed.call_count,
            '1 best_route_removed call for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False)])

    def test_c3_select_new_best_route_among_several(self):
        # When current best route is withdrawn, the new best route should be
        # selected among several routes
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 3 sources: A, B and C
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')
        worker_c = worker.Worker(mock.Mock(), 'worker.Worker-C')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B advertises a route2 for NLRI1. Route1 is better than Route2
        self._append_call("RE2")
        route2_nlri1b = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source C advertises a route3 for NLRI1. Route2 is better than Route3
        self._append_call("RE3")
        route3_nlri1c = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_c, t.NH1, 100)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B withdraws route2 for NLRI1
        self._append_call("RE5")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source C withdraws route3 for NLRI1
        self._append_call("RE6")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_c, t.NH1, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", "RE3",
                          "RE4", t.NBR, t.BRR, "RE5",
                          t.NBR, t.BRR, "RE6", t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            3, self.tracker_worker.new_best_route.call_count,
            '3 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route2_nlri1b.route_entry),
                           (t.NLRI1, route3_nlri1c.route_entry)])
        self.assertEqual(
            3, self.tracker_worker.best_route_removed.call_count,
            '3 best_route_removed calls for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False),
             (t.NLRI1, route2_nlri1b.route_entry, False),
             (t.NLRI1, route3_nlri1c.route_entry, True)])

    def test_d1_ecmp_routes(self):
        # ECMP routes are routes advertised by the same worker with the same
        # LP and different NH
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 1 source: A
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source A advertises a route2 for NLRI1. route2 is equal to route1
        # with compare_routes, but the next_hop are different
        self._append_call("RE2")
        route2_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH2, 100)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE3")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100)
        # Source A withdraws route2 for NLRI1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH2, 100)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", t.NBR,
                          "RE3", t.BRR, "RE4", t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route2_nlri1a.route_entry)])
        self.assertEqual(
            2, self.tracker_worker.best_route_removed.call_count,
            '2 best_route_removed calls for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False),
             (t.NLRI1, route2_nlri1a.route_entry, True)])

    def test_e1_replace_br_is_nbr(self):
        # Advertise a route that replaces the best route and becomes the new
        # best route
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 1 source: A
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 200)
        # Source A advertises a route2 for NLRI1. Route1 is better than Route2
        # BUT Route2 replaces Route1
        self._append_call("RE2")
        route2_nrli1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100, route1_nlri1a.route_entry)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", t.NBR, t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route2_nrli1a.route_entry)])
        self.assertEqual(
            1, self.tracker_worker.best_route_removed.call_count,
            '1 best_route_removed call for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False)])

    def test_e2_replace_br_is_not_nbr(self):
        # Advertise a route that replaces the best route but does not become
        # the new best route
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 2 sources : A and B
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B advertises a route2. Route1 is better than Route2
        self._append_call("RE2")
        route2_nrli1b = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source A advertises a route3 for NLRI1. Route3 replaces Route1.
        # Route2 is better than route3.
        self._append_call("RE3")
        route3_nrli1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 100, route1_nlri1a.route_entry)
        # Source B withdraws route2 for NLRI1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", "RE3", t.NBR,
                          t.BRR, "RE4", t.NBR, t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            3, self.tracker_worker.new_best_route.call_count,
            '3 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route2_nrli1b.route_entry),
                           (t.NLRI1, route3_nrli1a.route_entry)])
        self.assertEqual(
            2, self.tracker_worker.best_route_removed.call_count,
            '2 best_route_removed calls for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False),
             (t.NLRI1, route2_nrli1b.route_entry, False)])

    def test_e3_replace_br_is_not_nbr(self):
        # Advertise a route that replaces the best route but does not become
        # the new best route
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 3 sources: A, B and C
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')
        worker_c = worker.Worker(mock.Mock(), 'worker.Worker-C')

        # Source A advertises route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B advertises route2 for NLRI1 : route1 is better than route2
        self._append_call("RE2")
        route2_nlri1 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source C advertises also route2
        self._append_call("RE3")
        self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_c, t.NH1, 200)
        # Source A advertises route3 which replaces route1
        self._append_call("RE4")
        self._new_route_event(engine.RouteEvent.ADVERTISE, t.NLRI1,
                              [t.RT1, t.RT2], worker_a, t.NH1, 100,
                              route1_nlri1.route_entry)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", "RE3", "RE4", t.NBR, t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new best route call for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1.route_entry),
                           (t.NLRI1, route2_nlri1.route_entry)])
        self.assertEqual(
            1, self.tracker_worker.best_route_removed.call_count,
            '1 best_route_removed call for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1.route_entry)])

    def test_e4_not_replace_br(self):
        # Advertise a route that does not replaces the best route and becomes
        # the new best route when the best route is withdrawn
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 2 sources : A and B
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')

        # Source A advertises a route1 for NLRI1
        self._append_call("RE1")
        route1_nlri1a = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)
        # Source B advertises a route2. Route1 is better than Route2
        self._append_call("RE2")
        route2_nlri1b = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source B advertises a route3 for NLRI1. Route3 replaces Route2.
        # Route1 is better than Route3
        self._append_call("RE3")
        route3_nlri1b = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 100, route2_nlri1b.route_entry)
        # Source A withdraws route1 for NLRI1
        self._append_call("RE4")
        self._new_route_event(
            engine.RouteEvent.WITHDRAW, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = ["RE1", t.NBR, "RE2", "RE3", "RE4", t.NBR, t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self.assertEqual(
            2, self.tracker_worker.new_best_route.call_count,
            '2 new new_best_route calls for NLRI1')
        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route1_nlri1a.route_entry),
                           (t.NLRI1, route3_nlri1b.route_entry)])
        self.assertEqual(
            1, self.tracker_worker.best_route_removed.call_count,
            '1 best_route_removed call for NLRI1')
        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1_nlri1a.route_entry, False)])

    def test_e5_replace_br_is_nbr_equal(self):
        # Same as E3, but the route that replaces our current best compares
        # equally to the two initially less preferred routes, and becomes best
        # route with them
        self.tracker_worker.new_best_route = mock.Mock(
            side_effect=self._call_list(t.NBR))
        self.tracker_worker.best_route_removed = mock.Mock(
            side_effect=self._call_list(t.BRR))

        # 3 sources: A, B and C
        worker_a = worker.Worker(mock.Mock(), 'worker.Worker-A')
        worker_b = worker.Worker(mock.Mock(), 'worker.Worker-B')
        worker_c = worker.Worker(mock.Mock(), 'worker.Worker-C')

        # Source A advertises route1 for NLRI1
        route1 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 300)

        # We will only check events after this first one
        # to allow for a order-independent test after RE4
        del self.tracker_worker.new_best_route.call_args_list[:]

        # Source B advertises route2 for NLRI1 : route1 is better than route2
        self._append_call("RE2")
        route2 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_b, t.NH1, 200)
        # Source C advertises also route2
        self._append_call("RE3")
        route3 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_c, t.NH2, 200)
        # Source A advertises route3 which replaces route1
        self._append_call("RE4")
        route4 = self._new_route_event(engine.RouteEvent.ADVERTISE,
                                       t.NLRI1, [t.RT1, t.RT2],
                                       worker_a, t.NH3, 200,
                                       route1.route_entry)

        # Check calls and arguments list to new_best_route and
        # best_route_removed
        expected_calls = [t.NBR, "RE2", "RE3", "RE4",
                          t.NBR, t.NBR, t.NBR, t.BRR]
        self.assertEqual(expected_calls, self._calls, 'Wrong call sequence')

        self._check_calls(self.tracker_worker.new_best_route.call_args_list,
                          [(t.NLRI1, route2.route_entry),
                           (t.NLRI1, route3.route_entry),
                           (t.NLRI1, route4.route_entry)], False)

        self._check_calls(
            self.tracker_worker.best_route_removed.call_args_list,
            [(t.NLRI1, route1.route_entry, False)])

    def test_max_lp_is_best(self):
        worker_a = worker.Worker(mock.Mock(), 'worker')

        route_lp55 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 55).route_entry

        route_lp45 = self._new_route_event(
            engine.RouteEvent.ADVERTISE, t.NLRI1, [t.RT1, t.RT2],
            worker_a, t.NH1, 45).route_entry

        self.assertGreater(
            tracker_worker.compare_ecmp(mock.Mock(),
                                        route_lp55, route_lp45), 0)

        self.assertGreater(
            tracker_worker.compare_no_ecmp(mock.Mock(),
                                           route_lp55, route_lp45), 0)



# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import socket

from oslo_log import log as logging

from networking_bagpipe.bagpipe_bgp.common import log_decorator
from networking_bagpipe.bagpipe_bgp.common import looking_glass as lg
from networking_bagpipe.bagpipe_bgp.common import utils
from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.bagpipe_bgp.engine import worker


# Explanations on FilteredRoute

# We need to call self.best_route_removed for a route that was
# implicitly withdrawn, except if, in best_routes, there is
# a route that is the same with reference to what the
# best_route_removed callback will do:
# - if there is a route with same nexthop, same encap, same label
#   (other attributes make the route different)
#   => no call to best_route_removed
# - if there is a route with same nexthop, but a different label or
#   a different encap
#   => need to call best_route_removed
#
# Similarly, when multiple best routes exist for an entry
# (ECMP case), we don't want to call new_best_route multiple times
# uselessly for instance for two routes that would be 'same' wrt
# the actions done by the new_best_route callback. So we will do a
# call to self.new_best_route... *only* if the
# new_route is different from all current best routes.
#
# FilteredRoute is the class we use to identify a route
# that would be the 'same', based on the above.  For instance
# FilteredRoute removes information on the .source and
# only BGP attributes that are used by callbacks.
keep_attributes_default = [exa.Attribute.CODE.NEXT_HOP,
                           exa.Attribute.CODE.PMSI_TUNNEL,
                           exa.Attribute.CODE.MED,
                           exa.Attribute.CODE.EXTENDED_COMMUNITY,
                           exa.Attribute.CODE.LOCAL_PREF]


class FilteredRoute(engine.RouteEntry):

    def __init__(self, re, keep_attributes=None):
        if keep_attributes is None:
            keep_attributes = keep_attributes_default

        attributes = exa.Attributes()
        for (attribute_id, attribute) in re.attributes.items():
            if attribute_id in keep_attributes:
                attributes.add(attribute)

        super(FilteredRoute, self).__init__(re.nlri, None, attributes)


def equivalent_route_in_routes(function, route, routes):
    reference = function(route)
    for r in routes:
        if function(r) == reference:
            return True
    return False

# a compare_route callback has the following signature:
#
# def compare_route(self, route_a, route_b):
#     """
#     should return:
#      - an int>0 if route_a is better than route_b
#      - an int<0 if route_b is better than route_a
#      - else 0
#     """

# TODO(tmorin): both comparison should MAC Mobility if present


def compare_ecmp(worker, route_a, route_b):
    # makes a comparison based on LOCAL_PREF
    # if route has no LOCAL_PREF, use default value (100) as defined in exabgp
    l_a = route_a.attributes.get(exa.Attribute.CODE.LOCAL_PREF)
    l_a = l_a.localpref if l_a else 100
    l_b = route_b.attributes.get(exa.Attribute.CODE.LOCAL_PREF)
    l_b = l_b.localpref if l_b else 100
    return (l_a > l_b) - (l_b > l_a)


def compare_no_ecmp(worker, route_a, route_b):
    '''comparison operator without ECMP

    This compares the two routes in a consistent fashion, but two routes
    will never be considered of equal cost.
    The comparison is 'salted' so that two distinct VRFs (e.g. on two distinct
    bagpipe-bgp instances will not necessarily elect the same route as the
    best one).
    '''
    worker.log.trace("compare_no_ecmp used")
    base_comparison = compare_ecmp(worker, route_a, route_b)
    if base_comparison:
        return base_comparison

    salt = socket.gethostname() + worker.name
    hash_a = hash(salt + repr(route_a))
    hash_b = hash(salt + repr(route_b))
    cmp_hashes = (hash_a > hash_b) - (hash_b > hash_a)

    if cmp_hashes:
        return cmp_hashes

    return (route_a > route_b) - (route_b > route_a)


class TrackerWorker(worker.Worker, lg.LookingGlassLocalLogger,
                    metaclass=abc.ABCMeta):

    def __init__(self, bgp_manager, worker_name,
                 compare_routes=compare_no_ecmp):

        worker.Worker.__init__(self, bgp_manager, worker_name)
        lg.LookingGlassLocalLogger.__init__(self)

        # dict: entry -> list of routes:
        self.tracked_entry_2_routes = dict()
        # dict: entry -> set of best_routes:
        self.tracked_entry_2_best_routes = dict()

        self._compare_routes = compare_routes

    @log_decorator.log
    def _on_event(self, event):
        new_route = event.route_entry
        filtered_new_route = FilteredRoute(new_route)

        entry = self.route_to_tracked_entry(new_route)

        if entry is None:
            self.log.debug("Route not mapped to a tracked entry, ignoring: %s",
                           new_route)
            return

        self.log.trace("tracked_entry for this route: %s (type: %s)",
                       TrackerWorker._display_entry(entry), type(entry))

        self._dump_state()

        all_routes = self.tracked_entry_2_routes.setdefault(entry, [])

        self.log.debug("We currently have %d route%s for this entry",
                       len(all_routes), utils.plural(all_routes))

        if event.type == engine.RouteEvent.ADVERTISE:

            withdrawn_best_routes = []

            best_routes = self.tracked_entry_2_best_routes.get(entry)

            if best_routes is None:
                self.log.trace("We had no route for this entry (%s)")
                self.tracked_entry_2_best_routes[entry] = set([new_route])
                best_routes = set()
                self._call_new_best_route(entry, filtered_new_route)
            else:
                if event.replaced_route is not None:
                    self.log.trace("Will remove replaced route from all_routes"
                                   " and best_routes: %s",
                                   event.replaced_route)
                    try:
                        all_routes.remove(event.replaced_route)
                    except ValueError:
                        # we did not have any route for this entry
                        self.log.error("replaced_route is an entry for which "
                                       "we had no route ??? (bug ?)")

                    if event.replaced_route in best_routes:
                        self.log.trace(
                            "Removing replaced_route from best_routes")
                        best_routes.remove(event.replaced_route)
                        withdrawn_best_routes.append(event.replaced_route)
                    else:
                        self.log.trace("replaced_route is not in best_routes")
                        self.log.trace("best_routes: %s", best_routes)
                else:
                    self.log.trace("No replaced route to remove")

                call_new_best_route_4_all = False
                if len(best_routes) == 0:
                    self.log.trace("All best routes have been replaced")
                    self._recompute_best_routes(all_routes, best_routes)
                    if best_routes:
                        current_best = next(iter(best_routes))
                        self.log.trace("We'll need to call new_best_route for "
                                       "all our new best routes")
                        call_new_best_route_4_all = True
                    else:
                        current_best = None
                        call_new_best_route_4_all = False
                else:
                    # (if there is more than one route in the best routes, we
                    # take the first one)
                    current_best = next(iter(best_routes))

                    self.log.trace("Current best route: %s", current_best)

                    if new_route == current_best:
                        self.log.info("New route is a route we already had, "
                                      "nothing to do.")
                        # nothing to do
                        return

                # let's find if we need to update our best routes
                if current_best:
                    route_comparison = self._compare_routes(self, new_route,
                                                            current_best)
                else:
                    route_comparison = 1

                self.log.trace("route_comparison: %d", route_comparison)

                if route_comparison > 0:
                    # new_route is a strictly better route than any current
                    # one, discard all the current best routes
                    self.log.trace("Replacing all best routes with new one")
                    withdrawn_best_routes.extend(best_routes.copy())
                    best_routes.clear()
                    best_routes.add(new_route)
                    self._call_new_best_route(entry, filtered_new_route)
                    call_new_best_route_4_all = False
                elif route_comparison == 0:
                    # new_route is as good as the current ones
                    self.log.trace("Adding new_route to best_routes...")

                    if call_new_best_route_4_all:
                        self._call_new_best_route_for_routes(entry,
                                                             best_routes)

                    if not equivalent_route_in_routes(FilteredRoute,
                                                      new_route,
                                                      best_routes):
                        best_routes.add(new_route)
                        self.log.trace("Calling self.new_best_route since we "
                                       "yet had no such route in best routes")
                        self._call_new_best_route(entry, filtered_new_route)
                    else:
                        best_routes.add(new_route)
                        self.log.trace("Not calling new_best_route since we "
                                       "had received a similar route already")
                else:
                    self.log.trace("The route is no better than current "
                                   "best ones")

                    if call_new_best_route_4_all:
                        self._call_new_best_route_for_routes(entry,
                                                             best_routes)

            self.log.trace("Considering implicitly withdrawn best routes")
            for r in withdrawn_best_routes:
                self._selective_best_route_removed(entry,
                                                   r, best_routes,
                                                   False)

            # add the route to the list of routes for this entry
            self.log.trace("Adding route to all_routes for this entry")
            all_routes.append(new_route)

        else:  # RouteEvent.WITHDRAW

            withdrawn_route = new_route

            self.log.trace("Removing route from all_routes for this entry")

            # let's update known routes for this entry
            try:
                all_routes.remove(withdrawn_route)
            except ValueError:
                # we did not have any route for this entry
                self.log.error("Withdraw received for an entry for which we "
                               "had no route ??? (not supposed to happen)")

            # let's now update best routes
            best_routes = self.tracked_entry_2_best_routes.get(entry)

            if best_routes is None:
                # we did not have any route for this entry
                self.log.error("Withdraw received for an entry for which we "
                               "had no route: not supposed to happen!")
                return

            if withdrawn_route in best_routes:
                self.log.trace("The event received is about a route which "
                               "is among the best routes for this entry")
                # remove the route from best_routes
                best_routes.remove(withdrawn_route)

                withdrawn_route_is_last = False
                if len(best_routes) == 0:
                    # we don't have any best route left...
                    self._recompute_best_routes(all_routes, best_routes)

                    if len(best_routes) > 0:
                        self._call_new_best_route_for_routes(entry,
                                                             best_routes)
                    else:
                        self.log.trace("Cleanup all_routes and best_routes")
                        withdrawn_route_is_last = True
                        del self.tracked_entry_2_best_routes[entry]
                        del self.tracked_entry_2_routes[entry]

                self._selective_best_route_removed(entry,
                                                   withdrawn_route,
                                                   best_routes,
                                                   withdrawn_route_is_last)
            else:
                self.log.trace("The event received is not related to any "
                               "of the best routes for this entry")
                # no need to update our best route list
                pass

        self.log.info("We now have %d route%s for this entry.",
                      len(all_routes), utils.plural(all_routes))

        self._dump_state()

    def _recompute_best_routes(self, all_routes, best_routes):
        '''internal method to update best routes

        update best_routes to contain the best routes from all_routes, based
        on _compare_routes
        '''

        new_best_routes = []
        for route in all_routes:
            if len(new_best_routes) == 0:
                new_best_routes = [route]
                continue

            comparison = self._compare_routes(self, route, new_best_routes[0])
            if comparison > 0:
                new_best_routes = [route]
            elif comparison == 0:
                new_best_routes.append(route)

        best_routes.clear()
        best_routes.update(new_best_routes)

        self.log.trace("Recomputed new best routes: %s", best_routes)

    @log_decorator.log
    def _selective_best_route_removed(self, entry, withdrawn_route,
                                      filtered_best_routes,
                                      withdrawn_route_is_last):
        if not equivalent_route_in_routes(FilteredRoute,
                                          withdrawn_route,
                                          filtered_best_routes):
            self._call_best_route_removed(entry,
                                          FilteredRoute(withdrawn_route),
                                          withdrawn_route_is_last)
        else:
            self.log.trace("No need to call best_route_removed: %s",
                           FilteredRoute(withdrawn_route))

    def equivalent_route_in_best_routes(self, route, function):
        # This method checks if there is in the best routes for the same
        # tracked entry as 'route', at least one route r for which function(r)
        # is equal to function(route)
        return equivalent_route_in_routes(
            function,
            route,
            self.tracked_entry_2_best_routes.get(
                self.route_to_tracked_entry(route),
                []
            )
        )

    def synthesize_withdraw_all(self, afi, safi):
        for tracked_entry, routes in list(self.tracked_entry_2_routes.items()):
            self.log.trace("Synthesizing withdraws for all routes of %s with "
                           "AFI(%s)/SAFI(%s)", tracked_entry, afi, safi)
            for route in routes:
                if (route.nlri.afi, route.nlri.safi) == (afi, safi):
                    self._on_event(
                        engine.RouteEvent(engine.RouteEvent.WITHDRAW, route)
                    )

    @log_decorator.log
    def _call_new_best_route_for_routes(self, entry, routes):
        routes_no_dups = set([FilteredRoute(r) for r in routes])
        self.log.debug("   After filtering duplicates: %s", routes_no_dups)
        for route in routes_no_dups:
            self._call_new_best_route(entry, route)

    @log_decorator.log
    def _call_new_best_route(self, entry, new_route):
        try:
            self.new_best_route(entry, new_route)
        except Exception as e:
            self.log.error("Exception in <subclass>.new_best_route: %s", e)
            if self.log.isEnabledFor(logging.WARNING):
                self.log.exception("")

    @log_decorator.log
    def _call_best_route_removed(self, entry, old_route, last):
        try:
            self.best_route_removed(entry, old_route, last)
        except Exception as e:
            self.log.error("Exception in <subclass>.best_route_removed: %s",
                           e)
            if self.log.isEnabledFor(logging.WARNING):
                self.log.exception("")

    # Callbacks for subclasses ########################

    @abc.abstractmethod
    def route_to_tracked_entry(self, route):
        """Hook to control mapping from a route to a tracked entry

        This method is how the subclass maps a BGP route into an object that
        the TrackerWorker code will track.

        For instance, a VPN VRF is expected to keep track of IP prefixes;
        hence the route2tracked_entry code for a VRF could return the IP prefix
        in the VPNv4 route.
        The result will be that the TrackerWorker code will keep track, for a
        each prefix, of all the routes and of the best routes.
        """
        pass

    @abc.abstractmethod
    def new_best_route(self, entry, new_route):
        '''New Best Route hook

        A new best route has been advertized for this tracked entry
        '''
        pass

    @abc.abstractmethod
    def best_route_removed(self, entry, old_route, last):
        '''Best Route removed hook

        A route that was a best route for this tracked entry has been
        removed. The 'last' flag indicates if this was the last route
        for this tracked entry.
        '''
        pass

    # Debug support methods #########

    def _dump_state(self):
        if self.log.isEnabledFor(logging.TRACE):
            self.log.trace("--- tracked_entry_2_routes ---")
            for (entry, routes) in self.tracked_entry_2_routes.items():
                self.log.trace(
                    "  Entry: %s", TrackerWorker._display_entry(entry))
                for route in routes:
                    self.log.trace("    Route: %s", route)

            self.log.trace("--- tracked_entry_2_best_routes ---")
            for (entry, best_routes) in \
                    self.tracked_entry_2_best_routes.items():
                self.log.trace("  Entry: %s",
                               TrackerWorker._display_entry(entry))
                for route in best_routes:
                    self.log.trace("    Route: %s", route)

            self.log.trace("--- ---")

    @staticmethod
    def _display_entry(entry):
        if (isinstance(entry, tuple) and len(entry) > 0 and
                isinstance(entry[0], type)):
            return repr(tuple([entry[0].__name__] + list(entry[1:])))
        else:
            return repr(entry)

    # Looking glass ###########

    def get_lg_map(self):
        return {"received_routes": (lg.SUBTREE, self.get_lg_all_routes),
                "best_routes": (lg.SUBTREE, self.get_lg_best_routes)}

    def get_lg_all_routes(self, path_prefix):
        return self._get_lg_routes(path_prefix,
                                   self.tracked_entry_2_routes)

    def get_lg_best_routes(self, path_prefix):
        return self._get_lg_routes(path_prefix,
                                   self.tracked_entry_2_best_routes)

    def _get_lg_routes(self, path_prefix, route_dict):
        routes = {}
        for entry in route_dict.keys():
            entry_repr = self._display_entry(entry)
            routes[entry_repr] = [route.get_looking_glass_info(path_prefix)
                                  for route in route_dict[entry]]
        return routes



# Copyright 2017 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from oslo_config import cfg
import pecan
import socketserver
from wsgiref import simple_server


ROOT_CTRL = 'networking_bagpipe.bagpipe_bgp.api.controllers.RootController'


def setup_app(*args, **kwargs):
    config = {
        'server': {
            'port': cfg.CONF.API.port,
            'host': cfg.CONF.API.host,
        },
        'app': {
            'root': ROOT_CTRL,
        }
    }
    pecan_config = pecan.configuration.conf_from_dict(config)

    app = pecan.make_app(
        pecan_config.app.root,
        debug=False,
        force_canonical=False,
        guess_content_type_from_ext=True
    )

    return app


class ThreadedSimpleServer(socketserver.ThreadingMixIn,
                           simple_server.WSGIServer):
    pass


class PecanAPI(object):

    def __init__(self):

        app = setup_app()

        self.wsgi = simple_server.make_server(
            cfg.CONF.API.host,
            cfg.CONF.API.port,
            app,
            server_class=ThreadedSimpleServer
        )

    def run(self):
        self.wsgi.serve_forever()

    def stop(self):
        # call stop on RootController
        self.wsgi.get_app().application.root.stop()


def main():
    api = PecanAPI()
    api.run()


if __name__ == "__main__":
    main()



# Copyright (c) 2015 Orange.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as n_constants
from oslo_config import cfg
from oslo_log import log

from neutron.agent import securitygroups_rpc

from neutron.plugins.ml2.drivers import mech_agent

from neutron_lib.api.definitions import portbindings
from neutron_lib import constants as n_const


LOG = log.getLogger(__name__)


ml2_bagpipe_opts = [
    cfg.IntOpt('as_number', default=-1,
               help=("not used: bagpipe AS configuration for generation of "
                     "EVPN RTs must be done on neutron l2 agents"))
]

cfg.CONF.register_opts(ml2_bagpipe_opts, "ml2_bagpipe")


class BaGPipeMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """ML2 Mechanism driver for bagpipe-bgp

    This mechanism driver uses RPCs toward compute node agents to trigger
    the attachment of VM ports in E-VPN VPN instances.
    """

    def __init__(self):
        sg_enabled = securitygroups_rpc.is_firewall_enabled()
        super(BaGPipeMechanismDriver, self).__init__(
            n_const.AGENT_TYPE_LINUXBRIDGE,
            portbindings.VIF_TYPE_BRIDGE,
            {portbindings.CAP_PORT_FILTER: sg_enabled})

        if cfg.CONF.ml2_bagpipe.as_number != -1:
            raise Exception(
                "bagpipe AS configuration must be done on neutron l2 agents, "
                "in [ml2_bagpipe_extension]")

    def get_allowed_network_types(self, agent):
        return (agent['configurations'].get('tunnel_types', []) +
                [n_constants.TYPE_LOCAL, n_constants.TYPE_FLAT,
                 n_constants.TYPE_VLAN, n_constants.TYPE_VXLAN])

    def get_mappings(self, agent):
        return agent['configurations'].get('interface_mappings', {})



# Copyright 2014 Orange
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

.. module:: test_vpn_instance
   :synopsis: module with unit tests for bagpipe.bgp.engine.__init__

Validates the behavior of base objects of the engine.

Also validates that ExaBGP classes behave as expected by the code in
bagpipe.bgp.engine.__init__ .

"""

from testtools import TestCase

from networking_bagpipe.bagpipe_bgp import engine
from networking_bagpipe.bagpipe_bgp.engine import exa
from networking_bagpipe.tests.unit.bagpipe_bgp import base


TEST_RD = exa.RouteDistinguisher.fromElements("42.42.42.42", 5)


class TestEngineObjects(TestCase):

    def setUp(self):
        super(TestEngineObjects, self).setUp()

    def tearDown(self):
        super(TestEngineObjects, self).tearDown()

    # Tests on EVPN NLRIs

    def test_100_evpn_mac_hash_equal(self):
        # Two indistinct EVPN NLRI should
        # hash to the same value, and be equal

        nlri1 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("1.1.1.1"))

        nlri2 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("1.1.1.1"))

        self.assertEqual(hash(nlri1), hash(nlri2))
        self.assertEqual(nlri1, nlri2)

    def test_101_evpn_hash_equal_somefieldsvary(self):
        # Two EVPN MAC NLRIs differing by their ESI or label or RD,
        # or nexthop, but otherwise identical should hash to the same value,
        # and be equal

        nlri0 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("1.1.1.1"))

        # Esi
        nlri1 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(b''.join(bytes((1,))
                                             for _ in range(0, 10))),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("1.1.1.1"))

        # label
        nlri2 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([4444], True),
                            exa.IP.create("1.1.1.1"))

        # IP: different IPs, but same MACs: different route
        nlri3 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("2.2.2.2"))

        # with a next hop...
        nlri4 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("1.1.1.1"),
                            exa.IP.pton("10.10.10.10"))
        nlri5 = exa.EVPNMAC(TEST_RD,
                            exa.ESI(),
                            exa.EthernetTag(111),
                            exa.MAC("01:02:03:04:05:06"), 6 * 8,
                            exa.Labels([42], True),
                            exa.IP.create("1.1.1.1"),
                            exa.IP.pton("11.11.11.11"))

        self.assertEqual(hash(nlri0), hash(nlri1))
        self.assertEqual(hash(nlri0), hash(nlri2))
        self.assertEqual(hash(nlri0), hash(nlri4))
        self.assertEqual(nlri0, nlri1)
        self.assertEqual(nlri0, nlri2)
        self.assertEqual(nlri0, nlri4)
        self.assertEqual(nlri1, nlri2)
        self.assertEqual(nlri1, nlri4)
        self.assertEqual(nlri2, nlri4)
        self.assertEqual(nlri4, nlri5)

        self.assertNotEqual(hash(nlri0), hash(nlri3))
        self.assertNotEqual(nlri0, nlri3)
        self.assertNotEqual(nlri1, nlri3)
        self.assertNotEqual(nlri2, nlri3)
        self.assertNotEqual(nlri3, nlri4)

    # tests on attributes

    def test_4_same_nlri_distinct_attributes(self):
        # Two routes with same NLRI but distinct attributes should
        # not be equal

        atts1 = exa.Attributes()
        atts1.add(exa.LocalPreference(10))

        atts2 = exa.Attributes()
        atts2.add(exa.LocalPreference(20))

        entry1 = engine.RouteEntry(base.NLRI1, None, atts1)
        entry2 = engine.RouteEntry(base.NLRI1, None, atts2)

        self.assertNotEqual(entry1, entry2)

    def test_5_same_nlri_same_attributes(self):
        # Two routes with same NLRI but and same attributes should
        # hash to the same values and be equal.

        atts1 = exa.Attributes()
        atts1.add(exa.LocalPreference(10))

        atts2 = exa.Attributes()
        atts2.add(exa.LocalPreference(10))

        entry1 = engine.RouteEntry(base.NLRI1, None, atts1)
        entry2 = engine.RouteEntry(base.NLRI1, None, atts2)

        self.assertEqual(hash(entry1), hash(entry2))
        self.assertEqual(entry1, entry2)

    def test_6_same_nlri_same_attributes_order_multivalued(self):
        # Two routes with same NLRI but and same attributes should
        # hash to the same values and be equal, *even if* for a said
        # multivalued attributes, like extended community, the values
        # appear in a distinct order

        atts1 = exa.Attributes()
        ecoms1 = exa.ExtendedCommunities()
        ecoms1.communities.append(exa.RouteTarget(64512, 1))
        ecoms1.communities.append(exa.Encapsulation(
            exa.Encapsulation.Type.VXLAN))
        ecoms1.communities.append(exa.RouteTarget(64512, 2))
        atts1.add(ecoms1)

        atts2 = exa.Attributes()
        ecoms2 = exa.ExtendedCommunities()
        ecoms2.communities.append(exa.RouteTarget(64512, 2))
        ecoms2.communities.append(exa.RouteTarget(64512, 1))
        ecoms2.communities.append(exa.Encapsulation(
            exa.Encapsulation.Type.VXLAN))
        atts2.add(ecoms2)

        entry1 = engine.RouteEntry(base.NLRI1, None, atts1)
        entry2 = engine.RouteEntry(base.NLRI1, None, atts2)

        self.assertEqual(hash(entry1), hash(entry2))
        self.assertEqual(entry1, entry2)

    def test_8_route_entry_set_rts(self):
        atts = exa.Attributes()
        ecoms = exa.ExtendedCommunities()
        ecoms.communities.append(exa.RouteTarget(64512, 1))
        ecoms.communities.append(exa.RouteTarget(64512, 2))
        ecoms.communities.append(exa.Encapsulation(
            exa.Encapsulation.Type.VXLAN))
        atts.add(exa.LocalPreference(20))
        atts.add(ecoms)

        entry = engine.RouteEntry(base.NLRI1, None, atts)

        # check that the route_entry object has the RTs we wanted
        self.assertIn(exa.RouteTarget(64512, 1), entry.route_targets)
        self.assertIn(exa.RouteTarget(64512, 2), entry.route_targets)

        # modify the route targets
        entry.set_route_targets([exa.RouteTarget(64512, 3),
                                 exa.RouteTarget(64512, 1)])

        # check that the new RTs have replaced the old ones
        self.assertIn(exa.RouteTarget(64512, 1), entry.route_targets)
        self.assertIn(exa.RouteTarget(64512, 3), entry.route_targets)
        self.assertNotIn(exa.RouteTarget(64512, 2), entry.route_targets)

        # also need to check the RTs in the attributes
        ecoms = entry.attributes[
            exa.Attribute.CODE.EXTENDED_COMMUNITY].communities
        self.assertIn(exa.RouteTarget(64512, 1), ecoms)
        self.assertIn(exa.RouteTarget(64512, 3), ecoms)
        self.assertNotIn(exa.RouteTarget(64512, 2), ecoms)

        # check that other communities were preserved
        self.assertIn(exa.Encapsulation(exa.Encapsulation.Type.VXLAN), ecoms)

    def test_9_route_entry_rts_as_init_param(self):
        atts = exa.Attributes()
        ecoms = exa.ExtendedCommunities()
        ecoms.communities.append(exa.Encapsulation(
            exa.Encapsulation.Type.VXLAN))
        atts.add(exa.LocalPreference(20))
        atts.add(ecoms)

        rts = [exa.RouteTarget(64512, 1), exa.RouteTarget(64512, 2)]

        entry = engine.RouteEntry(base.NLRI1, rts, atts)

        self.assertIn(exa.RouteTarget(64512, 1), entry.route_targets)
        self.assertIn(exa.RouteTarget(64512, 2), entry.route_targets)

        ecoms = entry.attributes[
            exa.Attribute.CODE.EXTENDED_COMMUNITY].communities
        self.assertIn(exa.RouteTarget(64512, 1), ecoms)
        self.assertIn(exa.RouteTarget(64512, 2), ecoms)
        self.assertIn(exa.Encapsulation(exa.Encapsulation.Type.VXLAN), ecoms)

    def test_10_ecoms(self):
        ecoms1 = exa.ExtendedCommunities()
        ecoms1.communities.append(exa.Encapsulation(
            exa.Encapsulation.Type.VXLAN))
        atts1 = exa.Attributes()
        atts1.add(ecoms1)

        ecoms2 = exa.ExtendedCommunities()
        ecoms2.communities.append(exa.Encapsulation(
            exa.Encapsulation.Type.VXLAN))
        ecoms2.communities.append(exa.RouteTarget(64512, 1))
        atts2 = exa.Attributes()
        atts2.add(ecoms2)

        self.assertFalse(atts1.sameValuesAs(atts2))
        self.assertFalse(atts2.sameValuesAs(atts1))

    def test_11_rts(self):
        rt1a = exa.RouteTarget(64512, 1)
        rt1b = exa.RouteTarget(64512, 1)

        rt3 = exa.RouteTarget(64512, 2)
        rt4 = exa.RouteTarget(64513, 1)

        self.assertEqual(hash(rt1a), hash(rt1b))
        # self.assertEqual(hash(rt1a), hash(rt2))
        self.assertNotEqual(hash(rt1a), hash(rt3))
        self.assertNotEqual(hash(rt1a), hash(rt4))

        self.assertEqual(rt1a, rt1b)
        # self.assertEqual(rt1a, rt2)
        self.assertNotEqual(rt1a, rt3)
        self.assertNotEqual(rt1a, rt4)

        self.assertEqual(set([rt1a]), set([rt1b]))
        # self.assertEqual(set([rt1a]), set([rt2]))
        self.assertEqual(1, len(set([rt1a]).intersection(set([rt1b]))))
        # self.assertEqual(1, len(set([rt2]).intersection(set([rt1b]))))
